"""
kmayerb 
2020-11-12

Assist the user in finding quasi public meta-clonotypes
"""
import pandas as pd
import numpy as np
import os
import scipy.sparse
import warnings
from palmotif import compute_pal_motif, svg_logo
from tcrdist.summarize import _occurs_N_str, member_summ, filter_is, test_for_subsets
from progress.bar import IncrementalBar
from tcrdist.tree import _default_sampler, _default_sampler_olga, allele_01

__all__ = ['_neighbors_fixed_radius',
		   '_K_neighbors_fixed_radius',
		   '_neighbors_variable_radius',
		   '_K_neighbors_variable_radius',
		   '_neighbors_sparse_fixed_radius',
		   '_neighbors_sparse_variable_radius',
		   'make_motif_logo',
		   'make_motif_logo_from_index',
		   '_quasi_public_meta_clonotypes']

def _neighbors_fixed_radius(pwmat, radius):
	""" Returns the list of neighbor column indices if within the fixed radius """
	return [list(np.nonzero(pwmat[ii,]<=radius)[0]) for ii in range(pwmat.shape[0]) ]

def _K_neighbors_fixed_radius(pwmat, radius):
	""" Returns the number of neighbors (self-inclusive) if within the fixed radius"""
	return [len(list(np.nonzero(pwmat[ii,]<=radius)[0])) for ii in range(pwmat.shape[0]) ]

def _neighbors_variable_radius(pwmat, radius_list):
	""" Returns the list of neighbor column indices if within the fixed radius """
	return [list(np.nonzero(pwmat[ii,]<=radius)[0]) for ii, radius in zip(range(pwmat.shape[0]), radius_list) ]

def _K_neighbors_variable_radius(pwmat, radius_list):
	""" Returns the number of neighbors (self-inclusive) if within the fixed radius"""
	return [len(list(np.nonzero(pwmat[ii,]<=radius)[0])) for ii, radius in zip(range(pwmat.shape[0]), radius_list) ]

def _neighbors_sparse_fixed_radius(csrmat, radius):
	""" 
	Returns the list of neighbor column indices within the fixed radius

	Parameter
	--------- 
	csrmat : scipy.sparse.csr_matrix

	radius : int

	Return
	------
	list of lists of ints

	Example 
	-------

	>>> M = np.array([[-1,0,1,4],[1,-1,0,2],[0,10,-1,3],[0,0,2,-1]])
	>>> S = scipy.sparse.csr_matrix(M)
	>>> NN = _neighbors_sparse_fixed_radius(csrmat = S, radius = 4)
	>>> NN
	[[0, 2], [0, 1, 3], [2, 3], [2, 3]]
	>>> assert NN == [[0, 2], [0, 1, 3], [2, 3], [2, 3]]
	"""
	S = csrmat.copy()
	S[S > radius] = 0
	S.eliminate_zeros()
	return [list(x) for x in np.split(S.indices, S.indptr[1:-1])]


def _neighbors_sparse_variable_radius(csrmat, radius_list, maxd= 50):
	""" 
	Returns the list of neighbor column indices within the fixed radius

	Parameter
	--------- 
	csrmat : scipy.sparse.csr_matrix
	
	radius : int

	Return
	------
	list of lists of ints

	Example 
	-------
	>>> from tcrdist.public import _neighbors_sparse_variable_radius
	>>> import scipy.sparse
	>>> M = np.array([[-1,1,1,4],[1,-1,1,2],[1,10,-1,3],[5,1,2,-1]])
	>>> S = scipy.sparse.csr_matrix(M)
	>>> N = _neighbors_sparse_variable_radius(csrmat = S, radius_list = [1,1,20,1])
	>>> N
	[[0, 1, 2], [0, 1, 2], [0, 1, 2, 3], [1,3]]
	>>> assert N ==  [[0, 1, 2], [0, 1, 2], [0, 1, 2, 3], [1,3]]
	"""
	S = csrmat.copy()
	NN = list()
	for i,radius in enumerate(radius_list):
		if radius == 0:
			#row = np.asarray(S[i, :].todense())[0]
			#row[row == 0] = radius + 1
			#x = np.nonzero(row == -1)[0]
			ind = np.nonzero(S.data[S.indptr[i]:S.indptr[i+1]] == -1)[0]
			col_indices = S.indices[S.indptr[i]:S.indptr[i+1]][ind].tolist()
			NN.append(col_indices)
		else:
			ind = np.nonzero(S.data[S.indptr[i]:S.indptr[i+1]] <= radius)[0]
			col_indices = S.indices[S.indptr[i]:S.indptr[i+1]][ind].tolist()
			NN.append(col_indices)
	return NN 



class TCRpublic():
	"""

	Attributes

	tcrrep : tcrdist.repertoire.TCRrep

	clone_df : pd.DataFrame
		Clones information with standard tcrdist3 column names.
	pwmat : np.array
		Pairwise distances
	tcrsamper : tcrsampler.TCRsampler
		TCRSampler instance initialized with appropriate background
		set.
	cdr3_name : str
		Column name for amino acid CDR3 e.g., 'cdr3_d_aa'.
	v_gene_name : str
		Column name for TR[ABGD]V gene e.g., 'v_d_gene'.
	nr_filter : bool
		If True, sequqences with the exact same neighbors as another set will be
		dropped
	output_html_name : str
		Filename for the output html output.
	labels : list
		List of columns to display on html output beneath each logo plot. 
	fixed_radius : False
		If False, clone_df must have a column radius. 
		If True, argument radius will be used to define 
		maximum distance from centroid to neighboring TCR.
	radius : int or None
		Theshold distance (<=) for neighborhood membership.
		If int, then all centroids will be assigned the same
		radius. Alterntively radius can be provided for each
		centroid sequence by including radius as a numeric column
		in clone_df.
	query_str : str
		The string to include sequences in output. For instance
		'qpublic == True and K_neighbors > 3', implies that only 
		grouping of 4 or more TCRs from at leåast two individuals
		will be retained. Alternatively, 'nsubject > 1' or 
		'qpublic == True' could be used as true minimum requirements 
		for quasi-publicity.
	kargs_member_summ : dict
		kwargs 
	kargs_motif : dict
		kwargs for the motif genertation 
	"""
	def __init__(self, tcrrep, organism = None, chain = None, fixed_radius = True, output_html_name = "quasi_public_clones.html"):
		
		self.tcrrep = tcrrep
		
		if organism is None:
			self.organism = tcrrep.organism 
		else:
			assert organism in ['human', 'mouse'], "TCRPublic <organism> argument must be 'human' or 'mouse'"
			self.organism = organism 
		
		if chain is None:
			self.chain= tcrrep.chains[0]
		else:
			assert chain in ['alpha','beta', 'gamma','delta'], "TCRPublic <chain> argument must be 'alpha', 'beta', 'gamma' or 'delta'"
			self.chain = chain

		self.output_html_name = output_html_name

		self.sort_columns = ['nsubject','K_neighbors']
		self.sort_ascending = False
		
		# Get chain specific atributes
		self.pw_mat_str = {'alpha': 'pw_alpha',
						  'beta' : 'pw_beta', 
						  'gamma': 'pw_gamma',
						  'delta': 'pw_delta'}[self.chain]
		
		self.cdr3_name = {'alpha': 'cdr3_a_aa',
						  'beta' : 'cdr3_b_aa', 
						  'gamma': 'cdr3_g_aa',
						  'delta': 'cdr3_d_aa'}[self.chain]

		self.v_gene_name = {'alpha': 'v_a_gene',
						   'beta'  : 'v_b_gene', 
						   'gamma' : 'v_g_gene',
						   'delta' : 'v_d_gene'}[self.chain]

		self.j_gene_name = {'alpha': 'j_a_gene',
						   'beta'  : 'j_b_gene', 
						   'gamma' : 'j_g_gene',
						   'delta' : 'j_d_gene'}[self.chain]

		self.nr_filter = True
		
		self.output_html_name = output_html_name

		self.labels = ['clone_id',
						self.cdr3_name, 
						self.v_gene_name,
						self.j_gene_name,
						'radius',
						'neighbors',
						'K_neighbors',
						'nsubject',
						'qpublic',
						f'{self.cdr3_name}.summary', 
						f'{self.v_gene_name}.summary',
						f'{self.j_gene_name}.summary',
						f'{self.cdr3_name}.summary',
						'subject.summary']
		
		self.fixed_radius = fixed_radius
		
		self.radius = 18
		
		self.query_str = 'nsubject > 3'
		
		self.kargs_member_summ = {
			'key_col'   : 'neighbors', 
			'count_col' : 'count',
			'addl_cols' : ['subject'],
			'addl_n'    : 4}
		
		self.kargs_motif = {
			'pwmat_str'  : self.pw_mat_str,
			'cdr3_name'  : self.cdr3_name,
			'v_name'     : self.v_gene_name,
			'gene_names' : [self.v_gene_name, self.j_gene_name]}

		# Get the Default sampler
		try: 
			self.tcrsampler = _default_sampler_olga(organism = self.organism, chain = self.chain)()
		except KeyError:
			self.tcrsampler = _default_sampler(organism = self.organism, chain = self.chain)()

	def report(self):

		result = _quasi_public_meta_clonotypes(clone_df    = self.tcrrep.clone_df.copy(), 
									  pwmat       = getattr(self.tcrrep, self.pw_mat_str),
									  tcrsampler  = self.tcrsampler,
									  cdr3_name   = self.cdr3_name,
									  v_gene_name = self.v_gene_name,
									  nr_filter   = self.nr_filter,
									  output_html_name  = self.output_html_name, 
									  sort_columns = self.sort_columns,
									  sort_ascending = self.sort_ascending,
									  labels            = self.labels,
									  fixed_radius      = self.fixed_radius,
									  radius            = self.radius,
									  query_str         = self.query_str, 
									  kargs_member_summ = self.kargs_member_summ,
									  kargs_motif       = self.kargs_motif)
		
		print(f"WRITING: {self.output_html_name}")
		return result




def make_motif_logo(tcrsampler,
					clone_df,
					pwmat,
					centroid = 'CASSPDIEKYF',
					v_gene = 'TRBV7-9*01',
					radius = 24,
					pwmat_str = 'pw_delta',
					cdr3_name = 'cdr3_d_aa',
					v_name = 'v_d_gene',
					gene_names = ['v_d_gene','j_d_gene']):

	"""
	Make a motif from a tcrrep clone_df, pwmat, and a tcrsampler. 

	Parameters
	----------
	tcrsampler : tcrsamper.TCRsampler,
	clone_df : pd.DataFrame,
	pwmat : np.array,
	centroid : str
		e.g.,'CASSPDIEKYF',
	v_gene : str
		e.g. 'TRBV7-9*01',
	radius = int
		e.g., 26,
	pwmat_str : str
		e.g.,'pw_delta',
	cdr3_name : str
		e.g., 'cdr3_d_aa',
	v_name : str
		e.g., 'v_d_gene',
	gene_names : list
		eg., ['v_d_gene','j_d_gene']

	Returns 
	-------
	svg : str
	svg_raw : str

	Notes
	-----
	
	There is a safety first, efficiency loss involved 
	since we are relocating neighbors that 
	may already be know, but by looking 
	up the row index <irow> fisrst matching V,CDR3 this 
	function can be evoked without knowing 
	anything about the positions of 
	the neighbors ahead of time. This is particularly useful 
	since clone_df order is not stable after groupby 
	and deduplication. 
	"""
	irow = clone_df[(clone_df[cdr3_name] == centroid) & (clone_df[v_name ] == v_gene)].index[0]

	dfnode = clone_df[pd.Series(pwmat[irow , :]) <= radius].copy()
	dfnode[gene_names[0]] = dfnode[gene_names[0]].apply(lambda x : allele_01(x))
	dfnode[gene_names[1]] = dfnode[gene_names[1]].apply(lambda x : allele_01(x))

	gene_usage = dfnode.groupby(gene_names).size()
	
	sampled_rep = tcrsampler.sample( gene_usage.reset_index().to_dict('split')['data'],
					flatten = True, depth = 100)

	sampled_rep  = [x for x in sampled_rep if x is not None]

	motif, stat = compute_pal_motif(
					seqs = dfnode[cdr3_name],
					refs = sampled_rep,
					centroid = centroid)

	svg = svg_logo(motif, return_str= True)

	motif_raw, _ = compute_pal_motif(
				seqs     = dfnode[cdr3_name],
				centroid = centroid)

	svg_raw = svg_logo(motif_raw, return_str= True)

	return svg, svg_raw


def make_motif_logo_from_index(tcrsampler,
							   ind,
							   clone_df,
							   centroid,
							   cdr3_name = 'cdr3_b_aa',
							   v_name = 'v_b_gene',
							   gene_names = ['v_b_gene','j_b_gene']):

	"""
	make motif logo from a specific index
	"""
	dfnode = clone_df.iloc[ind,:].copy()

	dfnode[gene_names[0]] = dfnode[gene_names[0]].apply(lambda x : allele_01(x))
	dfnode[gene_names[1]] = dfnode[gene_names[1]].apply(lambda x : allele_01(x))

	gene_usage = dfnode.groupby(gene_names).size()
	
	sampled_rep = tcrsampler.sample( gene_usage.reset_index().to_dict('split')['data'],
					flatten = True, depth = 100)

	sampled_rep  = [x for x in sampled_rep if x is not None]

	motif, stat = compute_pal_motif(
					seqs = dfnode[cdr3_name],
					refs = sampled_rep,
					centroid = centroid)

	svg = svg_logo(motif, return_str= True)

	motif_raw, _ = compute_pal_motif(
				seqs     = dfnode[cdr3_name],
				centroid = centroid)

	svg_raw = svg_logo(motif_raw, return_str= True)

	return svg, svg_raw




def _quasi_public_meta_clonotypes(clone_df, 
								  pwmat,
								  tcrsampler, 
								  cdr3_name = 'cdr3_d_aa',
								  v_gene_name = 'v_d_gene',
								  nr_filter = True,
								  output_html_name = "quasi_public_clones.html",
								  sort_columns = ['nsubject','K_neighbors'],
								  sort_ascending = False,
								  labels = ['clone_id',
											'cdr3_d_aa', 
											'v_d_gene',
											'j_d_gene',
											'radius',
											'neighbors',
											'K_neighbors',
											#'cdr3s',
											'nsubject',
											'qpublic',
											'cdr3_d_aa.summary',
											'v_d_gene.summary',
											'j_d_gene.summary',
											'subject.summary'],
								  fixed_radius = False,
								  radius = None,
								  query_str = 'qpublic == True & K_neighbors > 1',
								  kargs_member_summ = {
									'key_col'   : 'neighbors', 
									'count_col' : 'count',
									'addl_cols' : ['subject'],
									'addl_n'    : 4},
								  kargs_motif = {
									'pwmat_str'  : 'pw_delta',
									'cdr3_name'  : 'cdr3_d_aa',
									'v_name'     : 'v_d_gene',
									'gene_names' : ['v_d_gene','j_d_gene']}):
	
	"""
	_quasi_public_meta_clonotypes


	Parameters
	----------
	clone_df : pd.DataFrame
		Clones information with standard tcrdist3 column names.
	pwmat : np.array
		Pairwise distances
	tcrsamper : tcrsampler.TCRsampler
		TCRSampler instance initialized with appropriate background
		set.
	cdr3_name : str
		Column name for amino acid CDR3 e.g., 'cdr3_d_aa'.
	v_gene_name : str
		Column name for TR[ABGD]V gene e.g., 'v_d_gene'.
	nr_filter : bool
		If True, sequqences with the exact same neighbors as another set will be
		dropped
	output_html_name : str
		Filename for the output html output.
	labels : list
		List of columns to display on html output beneath each logo plot. 
	fixed_radius : False
		If False, clone_df must have a column radius. 
		If True, argument radius will be used to define 
		maximum distance from centroid to neighboring TCR.
	radius : int or None
		Theshold distance (<=) for neighborhood membership.
		If int, then all centroids will be assigned the same
		radius. Alterntively radius can be provided for each
		centroid sequence by including radius as a numeric column
		in clone_df.
	query_str : str
		The string to include sequences in output. For instance
		'qpublic == True and K_neighbors > 3', implies that only 
		grouping of 4 or more TCRs from at leåast two individuals
		will be retained. Alternatively, 'nsubject > 1' or 
		'qpublic == True' could be used as true minimum requirements 
		for quasi-publicity.
	kargs_member_summ : dict
		kwargs 
	kargs_motif : dict
		kwargs for the motif genertation 

	Returns
	-------
	Returns DataFrames in a Dictionary. 
		nn_summary : pd.DataFrame 
			DataFrame matchign clone_df with summary measures added
		quasi_public_df pd.DataFrame
			Dataframe with only those rows that match the <query_str> and nr_filter.
	{'nn_summary': nn_summary : pd.DataFrame, 
	 'quasi_public_df': quasi_public_df : nn_summary : pd.DataFrame}

	Notes
	-----
	Importantly a html file is written displaying the quasi-public meta-clonotypes
	
	The easiest way to integrate this with existing nieghbor_diff 
	add 'neighbors' and 'K_neighbors' to the clone df. Other columns 
	could be added as well, and then displayed if added to the 
	lis of labels.

	nn_clone_df = pd.concat([tr.clone_df, ndif[['neighbors', 'K_neighbors','val_0','ct_0']] ], axis = 1)

	Examples
	--------
	"""
	if 'neighbors' not in clone_df.columns:
		if fixed_radius:
			clone_df['radius'] = radius
			clone_df['neighbors']   = _neighbors_fixed_radius(pwmat = pwmat, radius = radius)
		else:
			assert 'radius' in clone_df.columns, "if not using fixed_radius, the clone_df must have a numeric 'radius' columns"
			clone_df['neighbors']   = _neighbors_variable_radius(pwmat = pwmat, radius_list = clone_df.radius)

	if 'K_neighbors' not in clone_df.columns:
		if fixed_radius:
			clone_df['K_neighbors'] = _K_neighbors_fixed_radius(pwmat = pwmat, radius = radius)
		else: 
			clone_df['K_neighbors'] = _K_neighbors_variable_radius(pwmat = pwmat, radius_list = clone_df.radius)


	if 'nsubject' not in clone_df.columns:
		clone_df['nsubject']   = clone_df['neighbors'].\
			apply(lambda x: clone_df['subject'].iloc[x].nunique())

	if 'qpublic' not in clone_df.columns:
		clone_df['qpublic']     = clone_df['nsubject'].\
			apply(lambda x: x > 1)

	nn_summary = member_summ(res_df   = clone_df,
							clone_df  = clone_df,
							**kargs_member_summ)
	nn_summary = nn_summary.rename(columns = {k:f'{k}.summary' for k in nn_summary.columns})

	clone_df['cdr3s']    = clone_df['neighbors'].apply(lambda x: clone_df[cdr3_name].iloc[x].to_list())
	

	clone_df = pd.concat([clone_df, nn_summary], axis = 1)

	quasi_public_df = clone_df.query(query_str).\
		sort_values(sort_columns, ascending = sort_ascending).\
		reset_index(drop = True).\
		copy()
	
	if quasi_public_df.shape[0] == 0:
		raise ValueError("UNFORTUNATELY NO QUASI PUBLIC CLOONES WERE FOUND, CONSIDER YOUR QUERY STRINGENCY")

	quasi_public_df['unique_set'] = test_for_subsets(quasi_public_df['neighbors'])

	if nr_filter:
		quasi_public_df = filter_is(quasi_public_df, 'unique_set', 1).reset_index(drop = True)

	print(f"GENERATING {quasi_public_df.shape[0]} QUASI-PUBLIC MOTIFS SATISFYING {query_str}")
	bar = IncrementalBar('Processing', max = quasi_public_df.shape[0])
	svgs = list()
	svgs_raw = list()
	for i,r in quasi_public_df.iterrows():
		bar.next()
		centroid = r[cdr3_name]
		v_gene   = r[v_gene_name]
		svg, svg_raw = make_motif_logo( tcrsampler = tcrsampler, 
										pwmat = pwmat,
										clone_df = clone_df,
										centroid = centroid ,
										v_gene = v_gene ,
										radius = r['radius'],
										**kargs_motif)
		svgs.append(svg)
		svgs_raw.append(svg_raw)
	bar.next();bar.finish()

	quasi_public_df['svg'] = svgs
	quasi_public_df['svg_raw'] = svgs_raw

	def shrink(s):
		s = s.replace('height="100%"', 'height="20%"')
		s = s.replace('width="100%"', 'width="20%"')
		return s

	print(labels)

	with open(output_html_name, 'w') as output_handle:
		for i,r in quasi_public_df.iterrows():
			#import pdb; pdb.set_trace()
			svg, svg_raw = r['svg'],r['svg_raw']
			output_handle.write("<br></br>")
			output_handle.write(shrink(svg))
			output_handle.write(shrink(svg_raw))
			output_handle.write("<br></br>")
			output_handle.write(pd.DataFrame(r[labels]).transpose().to_html())
			output_handle.write("<br></br>")

	return {'nn_summary': nn_summary, 'quasi_public_df': quasi_public_df, 'clone_df': clone_df}



"""
example 17
"""

def test_example_report_public():
    """
    tcrdist3 is particularly useful for finding 
    what we term quasi-public meta-clonotypes, 
    collections of biochemically similar TCRs 
    recognizing the same peptide-MHC. 

    The easist way to define meta-clonotypes
    is to compute pairwise distances between 
    TCRs found in an antigen-enriched 
    subrepertoire, abbreviated below as 
    <aesr>
    """
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic
    
        # <aesr_fn> antigen-enriched subrepertoire
    fn = os.path.join('tcrdist', 'data','covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv')
        # <aesr_df> antigen-enriched subrepertoire
    df = pd.read_csv(fn)
        # <tr> TCR repertoire
    tr = TCRrep(
        cell_df = df[['cohort','subject','v_b_gene', 'j_b_gene','cdr3_b_aa']].copy(), 
        organism = 'human', 
        chains = ['beta'], 
        db_file = 'alphabeta_gammadelta_db.tsv', 
        compute_distances = True)

        # <tp> TCRpublic class for reporting publicities, fixed radius 18, 'nsubject > 3'
    tp = TCRpublic(
        tcrrep = tr, 
        output_html_name = "quasi_public_clones.html")

        # by calling, .report() an html report is made
    public = tp.report()
        
        # Also, the following datafarme are available
        # <clone_df> pd.DataFrame clone_df from tr.clone_df 
        # with neighbors and summary measures appended
    public['clone_df']
        # <nn_summary> pd.DataFrame with just summary measures
    public['nn_summary']
        # <quasi_public_df> Non-redundant groups of quasipublic clones
    public['quasi_public_df']



"""
breadth.py 

This module concerns the estimation of clonal breadth, whether it be at the 
pathogen, protein, or epitope level. Once meta-clonotype have been defined they 
can be used to search for biochemically similar TCRs in bulk repertoires 
that are likely to share antigen recognition. It is possible that a
single TCR clonotype may be conformant with multiple TCR meta-clonotypes, 
so an accurate estimate of clonal breadth must avoid double counting clonotypes.  

To estimate clonal breadth of antigen-associated TCRs within 
a bulk repertoires with (N) productive clonotypes and (M) total 
productive templates, we use a set of (X) previously defined
antigen-associated meta-clonotypes (defined as a (i) Centroid TRV,CDR3, 
(ii) TCR-specific RADIUS, (iii) MOTIF. 

1. Compute the TCRdist between each centroid TCRij for i {1...i...X} 
and all bulk clones {1...j..M} using rectangular search with the 
tcrdist.repertoires.TCRrep.compute_sparse_rect_distances(), producing 
a sparse distance matrix. 

2. Next perform a long-form tabulation that records the network formed between 
all meta-clonotype centroids and bulk sequences within the specified radius. 
This is performed with the function tcrdist.breadth.long_form_tabulation().

The network is represented as a Pandas DataFrame. Where centroid sequences
are recorded as "cdr3_b_aa", "v_b_gene", "j_b_gene" and the conformant sequence
in the bulk repertoire is "cdr3_b_aa_hit", 'v_b_gene_hit', 'j_b_gene_hit'. 
Crucially there is a column "MOTIF" which indicates whether the CDR3 of 
the hit sequence is conformant with the regular expression in the column 
"regex". 

3. The long-form Pandas DataFrame can then be used as input to the function 
tcrdist.breadth.estimate_breadth_and_depth(). The unit of analysis -- that is, 
whether breadth refers to pathogen, protein, or epitope specific breadth --
can be specified in with the argument <breadth_cols>. Crucially, when 
running tcrdist.breadth.long_form_tabulation() the argument 
<search_cols> must include a column indicating the association between 
a metac-clonotype and a particular 'protein' or 'epitope'  
e.g., ['tag', 'protein', 'epitope', cdr3_b_aa', 'v_b_gene', 'j_b_gene', 
'pgen','regex', 'radius']

An example is provided in the documentation. 
"""
import numpy as np
import os
import pandas as pd
import math
from tcrdist.public import _neighbors_sparse_variable_radius

def long_form_tabulation(
	clone_df1,
	clone_df2,
	csrmat, 
	search_cols = ['tag','cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'pgen','regex', 'radius']):
    '''
    Tabulate metaclonotype-discovered hits in 
    bulk data as a long-form Pandas DataFrame
 
    Parameters
    ----------
    clone_df1 : DataFrame
        Typically TCRrep.clone_df of metaclonotypes 
    clone_df2 : DataFame
        Typically TCRrep.clone_df of bulk sequences
    csrmat : scipy.sparse.csrmat
        Typically TCRrep.rw_beta
    search_cols: list
        Default ['tag','feature','cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'pgen','regex', 'radius']
 
    Returns 
    -------
    long_form_df : DataFrame
        Columns: 
        'tag'                  - Metaclonotype Target Name
        'feature_key'          - Metaclonotype Feature (CDR3X+TRVX+TRJX+22)
        'cdr3_b_aa'            - Metaclonotype CDR3
        'v_b_gene'             - Metaclonotype V
        'j_b_gene'             - Metaclonotype J
        'pgen'.                - Metaclonotype pgen
        'regex'                - Metaclonotype regex
        'radius'               - Metaclonotype radius
        'cdr3_b_aa_hit'        - Discovered match CDR3
        'v_b_gene_hit'         - Discovered match CDR3
        'j_b_gene_hit'         - Discovered match CDR3
        'productive_frequency' - Discovered match sample Frequency
        'count'                - Discovered match sample template count
        'dist'                 - TCRdist to match
        'MOTIF'                - regex match (boolean)
    '''
    from tcrdist.public import _neighbors_sparse_variable_radius
    import re

    clone_df1['RADIUS'] = \
        _neighbors_sparse_variable_radius(
            csrmat      = csrmat, 
            radius_list = clone_df1.radius.to_list())
 
    # Tidy Long Form DataFrame to Record All Hits
    long_form = list()
    for i,r in clone_df1.iterrows():
        if len(r['RADIUS']) > 0:
            hit_features = clone_df2.iloc[r['RADIUS'],:]
            hit_features = hit_features[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'productive_frequency', 'count']]
            
            dists = csrmat[i,r['RADIUS']].toarray()[0].tolist() 
            hit_features['dist'] = dists
 
            search_features = r[search_cols]
            feature_key = search_features['cdr3_b_aa'] + '+' + search_features['v_b_gene'] + '+' + search_features['j_b_gene'] + '+' + str(search_features['radius'])
            
            search_features = search_features[search_cols]
            search_features['feature_key'] = feature_key
 
            hit_features['feature_key']    = feature_key
            
            dx = pd.DataFrame(search_features).\
            transpose().\
            merge( hit_features, how = 'left', left_on = 'feature_key', right_on = 'feature_key', suffixes = ['','_hit'])
            
            long_form.append(dx)
 
    long_form_df = pd.concat(long_form)
 
    def _re(regex, cdr3_b_aa_hit):
        return re.search(pattern = regex, string = cdr3_b_aa_hit) is not None

    long_form_df['MOTIF'] = long_form_df.apply(lambda x: _re(x['regex'], x['cdr3_b_aa_hit']), axis=1)
    
    return(long_form_df)


def estimate_breadth_and_depth(df,
    N,
    M, 
    breadth_cols = ['protein'], 
    clonal_cols = ['cdr3_b_aa_hit', 'v_b_gene_hit', 'j_b_gene_hit'],
    motif = True, 
    exact = False):
    """
    Computes Breadth and Depth of Pathogen, Protein or Epitope specific 
    clones.

	Parameters 
	----------
    df : DataFrame
        The result of long_from_tabulation()
    N : int
        number of unique productive TCR clonotypes 
    M : int 
        number of total sequenced productive TCRs 
    breadth_cols : list 
        ['protein'], the columns over which to compute breadth. 
        The Default ['protein'] will compute breadth for each 
        protein. If None, the breadth estimate will be based on 
        all clones specific to all proteins and epitopes. That 
        is it likely be a pathogen specific breadth.
    clonal_cols : list
        A list of columns that define what is a clone.
        to avoid
    motif : bool 
        If True, df.query("MOTIF == True") is applied, such that only 
        clones that matched a MOTIF are considered in an estimate.
    exact : bool 
        If true df.query("dist <= 0") is applied, such that only clones that 
        exactly matched a search centroid are considered. You probably don't 
        want ot use this except for comparisons sake. (Note that perfect matches
        TCRdist = 0, are designated as 0 in sparse format) 

	Returns 
	-------


    Notes
    -----
    This function is called after the function long_form_tabulation()
    and takes its output as input <df>, N and M can be extracted from 
    N =tr_bulk.clone_df.shape[0] , 
    M= tr_bulk.clone_df['count'].sum()))

    For protein specific breadth, input long-form DataFrame must contain 
    column specifying protein and argument <breadth_cols> should be 
    ['protein'. Similarly for epitope specific breadth <breadth_cols> should be 
    ['protein', 'epitope']. Setting the argument  <breadth_cols> to None
    will result in breadth and depth estimates based on the full set 
    of hits.

    Examples 
    --------

    lf_df = long_form_tabulation(
        clone_df1 = tr_search.clone_df,
        clone_df2 = tr_bulk.clone_df, 
        csrmat = tr_search.rw_beta, 
        search_cols = ['tag', 'protein','protein_coordinate','cdr3_b_aa',
                       'v_b_gene', 'j_b_gene', 'pgen','regex', 'radius'])

    # PROTEIN-specific breadth, RADIUS+MOTIF conformant clones 
    estimate_breadth_and_depth(df=lf_df, 
        breadth_cols = ['protein'], 
        N =tr_bulk.clone_df.shape[0] , 
        M= tr_bulk.clone_df['count'].sum(), 
        motif = True,
        exact = False)

    # PROTEIN-specific breadth, RADIUS conformant clones 
    estimate_breadth_and_depth(df=lf_df, 
        breadth_cols = ['protein'], 
        N =tr_bulk.clone_df.shape[0] , 
        M= tr_bulk.clone_df['count'].sum(), 
        motif = False,
        exact = False))\

    # PROTEIN-specific breadth, EXACT MATCHING clones 
    print(estimate_breadth_and_depth(df=lf_df, 
        breadth_cols = ['protein'], 
        N =tr_bulk.clone_df.shape[0] , 
        M= tr_bulk.clone_df['count'].sum(), 
        motif = True, 
        exact = True)
    
    # PATHOGEN-specific breadth, RADIUS+MOTIF conformant clones 
    print(estimate_breadth_and_depth(df=lf_df, 
        breadth_cols = None, 
        N =tr_bulk.clone_df.shape[0] , 
        M= tr_bulk.clone_df['count'].sum(), 
        motif = True, 
    exact = False))
    """
    if motif:
        df = df.query("MOTIF == True")
    if exact: 
        df = df.query("dist <= 0")

    if breadth_cols is None:
        print("No columns specified, breadth estimates are based on all results")

    # To avoid double counting we only take the first instance of 
    # each unique CDR3,V,G clonotype.
    df_unique = df.\
        sort_values(['tag','dist'], ascending=True).\
        groupby(by = clonal_cols).\
        first()
    # Next we compute the log2(1 + count) for each unique hit, note zeros evaluate to log2(1) = 0
    # and need not be calculated in the sum
    df_unique['log2_count'] = df_unique['count'].apply(lambda x : np.log2(1+x))
    df_unique['clone_count'] = 1 
    # Next we groupby at the level of interest
    if breadth_cols is None:
        df_unique['scope'] = "all eptopes"
        df_report = df_unique.groupby('scope')[['productive_frequency', 'clone_count', 'count','log2_count']].sum()
    else:
        df_report = df_unique.groupby(breadth_cols)[['productive_frequency', 'clone_count', 'count','log2_count']].sum()
    
    df_report['breadth'] = df_report['clone_count'].apply(lambda x : x / N)
    df_report['depth'] = df_report['log2_count'].apply(lambda x : x - np.log2(M))
    df_report['1_over_N'] = 1/N
    df_report['1_over_clone_count'] = 1/df_report['clone_count']
    df_report['SE_breadth'] = df_report['breadth'] * np.sqrt(df_report['1_over_N']+df_report['1_over_clone_count'])
    
    return df_report.sort_values('breadth', ascending= False)



def get_safe_chunk(search_clones, bulk_clones,target = 10**7):
	"""
	This function help pick a chunk size that prevents excessive memory use,
	With two CPU, 10*7 should keep total overall memory demand below 1GB
	"""
	ideal_divisor = (search_clones * bulk_clones) / target
	if ideal_divisor < 1:
		ideal_chunk_size = search_clones
		print(ideal_chunk_size)
	else:
		ideal_chunk_size = math.ceil((search_clones)/ ideal_divisor)
		print(ideal_chunk_size)
	return ideal_chunk_size




import pytest
from tcrdist.repertoire import TCRrep
import pandas as pd
import numpy as np


"""
This testing modul is for tcrdist.repertoire
"""

"""
INTEGRATION TESTS
"""
def test_gamma_delta_all_steps():
    """
    Test that correctly formated gamma delta data can be 
    fully processed during TCRrep initialization
    """
    df = pd.read_csv("sant.csv")
    tr = TCRrep(
    cell_df = df, 
    organism = "human", 
    chains = ['gamma', "delta"],
    infer_cdrs     = True,
    deduplicate    = True,
    use_defaults   = True,
    store_all_cdr  = True,
    compute_distances = True,
    cpus    = 1, 
    db_file = 'alphabeta_gammadelta_db.tsv')
    assert isinstance(tr.pw_gamma, np.ndarray)
    assert isinstance(tr.pw_delta, np.ndarray)

def test_alpha_beta_all_steps():
    """
    Test that correctly formated alpha beta data can be 
    fully processed during TCRrep initialization
    """
    df = pd.read_csv("dash.csv")
    tr = TCRrep(
    cell_df = df, 
    organism = "mouse", 
    chains = ['alpha', "beta"],
    infer_cdrs     = True,
    deduplicate    = True,
    use_defaults   = True,
    store_all_cdr  = True,
    compute_distances = True,
    cpus    = 1, 
    db_file = 'alphabeta_gammadelta_db.tsv')

    assert isinstance(tr.pw_alpha, np.ndarray)
    assert isinstance(tr.pw_beta, np.ndarray)

def test_alpha_beta_all_steps_imgt_aligned_FALSE():
    """
    Test that correctly formated alpha beta data can be 
    fully processed during TCRrep initialization
    """
    df = pd.read_csv("dash.csv")
    tr = TCRrep(
    cell_df = df, 
    organism = "mouse", 
    chains = ['alpha', "beta"],
    infer_cdrs     = True,
    imgt_aligned   = False,
    deduplicate    = True,
    use_defaults   = True,
    store_all_cdr  = True,
    compute_distances = True,
    cpus    = 1, 
    db_file = 'alphabeta_gammadelta_db.tsv')

    assert isinstance(tr.pw_alpha, np.ndarray)
    assert isinstance(tr.pw_beta, np.ndarray)

def test_gamma_delta_manually_step_by_step():
    """
    Test that the user can go through the automatic 
    steps, manually, one-by-one for well formatted
    gamma delta data.
    """
    df = pd.read_csv("sant.csv")
    tr = TCRrep(
    cell_df = df, 
    organism = "human", 
    chains = ['gamma', "delta"],
    imgt_aligned   = False,
    infer_cdrs     = False,
    infer_index_cols = False,
    deduplicate    = False,
    use_defaults   = False,
    store_all_cdr  = False,
    compute_distances = False,
    cpus    = 1, 
    db_file = 'alphabeta_gammadelta_db.tsv')

    tr.infer_cdrs_from_v_gene(chain = "gamma")
    tr.infer_cdrs_from_v_gene(chain = "delta")
    tr.infer_index_cols()
    tr.show_incomplete()
    tr.deduplicate()
    tr._initialize_chain_specific_attributes()
    tr.stora_all_cdr = True
    tr.compute_distances()
    tr.pw_gamma

def test_alpha_beta_manually_step_by_step():
    """
    Test that the user can go through the automatic 
    steps, manually, one-by-one for well formatted
    alpha beta data.
    """
    df = pd.read_csv("dash.csv")
    tr = TCRrep(
    cell_df = df, 
    organism = "mouse", 
    chains = ['alpha', "beta"],
    imgt_aligned   = False,
    infer_cdrs     = False,
    infer_index_cols = False,
    deduplicate    = False,
    use_defaults   = False,
    store_all_cdr  = False,
    compute_distances = False,
    cpus    = 1, 
    db_file = 'alphabeta_gammadelta_db.tsv')

    tr.infer_cdrs_from_v_gene(chain = "alpha")
    tr.infer_cdrs_from_v_gene(chain = "beta")
    tr.infer_index_cols()
    
    tr.deduplicate()
    tr._initialize_chain_specific_attributes()
    tr.stora_all_cdr = True
    tr.compute_distances()
    tr.pw_beta
   

"""
NULL INPUTS - Test that object can be created without any real inputs
"""
def test_null_import():
    """
    Null Import - Test that we can import with no inputs and everything set to false
    """
    with pytest.warns(UserWarning):
        tr = TCRrep(
            organism          = "mouse",
            chains            = ["alpha", "beta"],
            cell_df           = None,
            clone_df          = None,
            imgt_aligned      = False,
            infer_cdrs        = False,
            infer_index_cols  = False,
            deduplicate       = False,
            use_defaults      = False,
            store_all_cdr     = False,
            compute_distances = False,
            cpus              = 1, 
            db_file           = 'alphabeta_gammadelta_db.tsv')
    assert isinstance(tr.chains, list) 

def test_null_import_defaults_left_blank():
    with pytest.warns(UserWarning):
        tr = TCRrep(
            cell_df           = None,
            clone_df          = None,
            imgt_aligned      = False,
            infer_cdrs        = False,
            infer_index_cols  = False,
            deduplicate       = False,
            use_defaults      = False,
            store_all_cdr     = False,
            compute_distances = False)
    assert isinstance(tr.chains, list) 


def test_null_import_gamma_delta():
    with pytest.warns(UserWarning):
        tr = TCRrep(
            organism          = "human",
            chains            = ["gamma", "delta"],
            cell_df           = None,
            clone_df          = None,
            imgt_aligned      = False,
            infer_cdrs        = False,
            infer_index_cols  = False,
            deduplicate       = False,
            use_defaults      = False,
            store_all_cdr     = False,
            compute_distances = False,
            cpus              = 1, 
            db_file           = 'alphabeta_gammadelta_db.tsv')
    assert isinstance(tr.chains, list) 



"""
VALIDATION TESTS
"""
def test_validate_organism():
    """Test that incorrect organisms raises ValueError"""
    with pytest.raises(ValueError) as info:
        tr = TCRrep(
            organism          = "hamster",
            chains            = ["alpha", "beta"])
    assert str(info.value) == "organism must be 'mouse', 'human', or 'rhesus'"

def test_validate_chains():
    """Test that incorrect chain raise causes ValueError"""
    with pytest.raises(ValueError) as info:
        tr = TCRrep(
            organism          = "mouse",
            chains            = ["ALPHA", "beta"])
    assert str(info.value) == "TCRrep chains arg can be one or more of the following ['alpha', 'beta', 'gamma', 'delta'] case-sensitive"


def test_validate_imgt_aligned():
    """Test that incorrect chain raise causes ValueError"""
    with pytest.raises(ValueError) as info:
        tr = TCRrep(
            organism          = "mouse",
            chains            = ["alpha", "beta"], 
            imgt_aligned      = "align")
    assert str(info.value) == "TCRrep imgt_aligned argument must be a boolean"

def test_validate_cell_df_is_pandas_dataframe():
    with pytest.warns(UserWarning):
        tr = TCRrep(
            cell_df           = np.zeros(1),
            organism          = "mouse",
            chains            = ["alpha", "beta"],
            clone_df          = None,
            imgt_aligned      = False,
            infer_cdrs        = False,
            infer_index_cols  = False,
            deduplicate       = False,
            use_defaults      = False,
            store_all_cdr     = False,
            compute_distances = False,
            cpus              = 1, 
            db_file           = 'alphabeta_gammadelta_db.tsv')

def test_warn_on_bad_db_file():
    """Test that incorrect chain raise causes ValueError"""
    with pytest.warns(UserWarning):
        tr = TCRrep(
            organism          = "human",
            chains            = ["gamma", "delta"],
            cell_df           = None,
            clone_df          = None,
            imgt_aligned      = False,
            infer_all_genes   = False,
            infer_cdrs        = False,
            infer_index_cols  = False,
            deduplicate       = False,
            use_defaults      = False,
            store_all_cdr     = False,
            compute_distances = False,
            cpus              = 1, 
            db_file           = 'WRONGalphabeta_gammadelta_db.tsv')






def test_example_variable_radius():
    """
    Instead of enforcing a fixed radius, 
    use a radius specific to each
    centroid, specified in an additional 
    column.
    """
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic
    
    fn = os.path.join(
        'tcrdist',
        'data',
        'covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.radius.csv')

    df = pd.read_csv(fn)

    tr = TCRrep(cell_df = df[['cohort','subject','v_b_gene', 'j_b_gene','cdr3_b_aa', 'radius']], 
                organism = "human", 
                chains = ["beta"])

    tp = TCRpublic(
        tcrrep = tr, 
        output_html_name = "quasi_public_clones3.html")

    # set to True, if we want a universal radius
    tp.fixed_radius = False
    # must then specify maximum distance for finding similar TCRs
    tp.radius = None
    # set criteria for being quasi-public
    tp.query_str = 'nsubject > 5'
    # Add additional columns to be summarized in the report
    tp.kargs_member_summ['addl_cols'] = ['subject', 'cohort']
    # Add cohort.summary to the labels column so it shows up in the report
    tp.labels.append("cohort.summary")
    tp.labels.append("cdr3s")
    # Change number of subjects to display
    tp.kargs_member_summ['addl_n'] = 10
    # by calling, .report() an html report is made
    public = tp.report()













    # # <aesr_fn> antigen-enriched subrepertoire
    # fn = os.path.join(
    #     'tcrdist',
    #     'data',
    #     'covid19',
    #     'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv')

    # df = pd.read_csv(fn)
    
    # df = df[['subject',
    #     'cohort',
    #     'v_b_gene',
    #     'j_b_gene',
    #     'cdr3_b_aa']]

    # fn2 = os.path.join(
    #     'tcrdist',
    #     'data',
    #     'covid19',
    #     'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv.bE5ctrl.centers.csv')
    # df2 = pd.read_csv(fn2)
    # df2 = df2[[
    #     'cdr3_b_aa',
    #     'v_b_gene',
    #     'j_b_gene',
    #     'pgen', 
    #     'max_radi',
    #     'chi2joint']].\
    #     rename(columns = {'max_radi':'radius'}).\
    #     groupby(['v_b_gene','j_b_gene','cdr3_b_aa']).\
    #     first().reset_index(drop = False)
    # outfn = os.path.join(
    #     'tcrdist',
    #     'data',
    #     'covid19',
    #     'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.radius.csv')
    # df.merge(df2, how = "left", on = [ 'cdr3_b_aa','v_b_gene','j_b_gene']).to_csv(outfn, index = False)




"""
tcrdist/neigbhors.py

This module contas 
"""

import numpy as np 
import pandas as pd
import parmap
import scipy.sparse
import warnings
#from tcrdist.repertoire import TCRrep
from tcrdist.regex import _index_to_regex_str, _index_to_seqs, _multi_regex 
from scipy.stats import chi2_contingency



def compute_odds_ratio(pos, neg, bpos, bneg, ps = 1):
    """
    pos : int or float
        number of positve instances in sample
    neg : int or float
        number of negative instances in sample
    bpos : int or float 
        number of positve instances in background population
    bneg : int of float 
        number of negative instances in background population
    ps : int or float
        psuedocount to avoid infinite odds ratio when zeros are possible in the denominator

    Notes
    -----
    An odds ratio (OR) is a statistic that quantifies the strength of the association between two events, A and B. The odds ratio is defined as the ratio of the odds of A in the presence of B and the odds of A in the absence of B

    Examples
    --------    
    >>> compute_odds_ratio(10, 100, 10, 1000, ps = 0)
    10.0
    >>> compute_odds_ratio(10, 100, 10, 1000, ps = 1)
    9.91089108910891
    """
    odds_ratio = ((ps+pos)/(ps+neg))/ ((ps+bpos)/(ps+bneg))
    return odds_ratio

def compute_rate(pos,neg, ps = 1):
    return ((pos + ps) / (pos + neg + 2*ps))

def _todense_row(pw, maxd):
    """Make a sparse distance matrix dense, dealing with zeros as > maxd and -1 as true zero"""
    pw = np.asarray(pw.todense())
    pw[pw == 0] = maxd + 1
    pw[pw == -1] = 0
    return pw

#      888                                          888                                           888 d8b          888                                                                    888  .d888 
#      888                                          888                                           888 Y8P          888                                                                    888 d88P"  
#      888                                          888                                           888              888                                                                    888 888    
#  .d88888  .d88b.  88888b.   .d88b.  88888b.   .d88888 .d8888b        .d88b.  88888b.        .d88888 888 .d8888b  888888  8888b.  88888b.   .d8888b .d88b.          .d88b.   .d8888b .d88888 888888 
# d88" 888 d8P  Y8b 888 "88b d8P  Y8b 888 "88b d88" 888 88K           d88""88b 888 "88b      d88" 888 888 88K      888        "88b 888 "88b d88P"   d8P  Y8b        d8P  Y8b d88P"   d88" 888 888    
# 888  888 88888888 888  888 88888888 888  888 888  888 "Y8888b.      888  888 888  888      888  888 888 "Y8888b. 888    .d888888 888  888 888     88888888        88888888 888     888  888 888    
# Y88b 888 Y8b.     888 d88P Y8b.     888  888 Y88b 888      X88      Y88..88P 888  888      Y88b 888 888      X88 Y88b.  888  888 888  888 Y88b.   Y8b.            Y8b.     Y88b.   Y88b 888 888    
#  "Y88888  "Y8888  88888P"   "Y8888  888  888  "Y88888  88888P'       "Y88P"  888  888       "Y88888 888  88888P'  "Y888 "Y888888 888  888  "Y8888P "Y8888 88888888 "Y8888   "Y8888P "Y88888 888    
#                   888                                                                                                                                                                              
#                   888                                                                                                                                                                              
#                   888   

def bkgd_cntl_nn2( tr, 
                   tr_background,
                   weights = None,
                   ctrl_bkgd = 10**-5, 
                   col = 'cdr3_b_aa',
                   add_cols = ['v_b_gene', 'j_b_gene'],
                   pw_mat_str = 'pw_beta',
                   rw_mat_str = 'rw_beta',
                   ncpus = 4,
                   include_seq_info= True,
                   thresholds = [x for x in range(0,50,2)],
                   generate_regex = True,
                   test_regex = True,
                   beta_re = 1,
                   beta_dist = 1,
                   forced_max_radius = None):   

    """
    bkgd_cntl_nn2 stands for background controlled nearest neighbors. 

    tr : tcrdist.repertoire.TCRrep
        TCRrep instance with clone_df of target data
    tr_background : tcrdist.repertoire.TCRrep
        TCRrep instance with clone_df of bulk data
    ctrl_bkgd : float
        Default is 2*10**-5, the acceptable level of background neighbror 
    col : str
        Default is cdr3_b_aa', the column containing CDR3 string
    add_cols  : list
        Extra columns from clone_df to include in centers_df['v_b_gene', 'j_b_gene'],
    ncpus : int
        e.g., 4, passed to pm_processes when using parmap   
    include_seq_info : bool 
        If True, returned DataFrame <centers_df> will include ['target_neighbors', 'target_seqs',
        'background_neighbors','background_seqs','background_v', 'background_j']
        as columns in centers_df DataFrame returned by this function. 
        This allows for inspection of sequences found in both the antigen enriched repertoire and supplied
        background.
    thresholds : list
        Default is [x for x in range(0,50,2)] indicating tcrdist thresholds to compute 
        a ecdf over.
    generate_regex : bool
        if True, generate a regex pattern capturing most if not all of 
        sequences in the neigbhorhood
    test_regex = True
        if True, test regex against both CDR3 in target and background
    beta_re : int or float
        for joint chi2, weight for regex based portion
    beta_dist : int of float
        for joint chi2, weight for distance based portion
    forced_max_radius : int or None
        if not None, radius cannot exceed this amount
    
    Returns 
    -------
    centers_df: DataFrame
    
    Notes
    -----

    The logic behind this function is that we want to find radi that 
    are appropriate for each TCR in a epitope-specific target set 
    that control the frequency of finding neighbors in a background set. 
    For instance, a TCR with common V gene usage and a high probabililty 
    of generation CDR3, will potentially have many nieghbors in a 
    pre-selection background set, whereas a TCR with rare gene usage 
    and a low probability of generation CDR3 can acomodate a larger 
    neighborhood radi. 

    The result dataframe can be sorted to find TCRs that anchor 
    a nieghborhood of TCRs including the most epitope specific
    nieghbors. We hypothesis that these TCRs are more likely to serve as a useful 
    biomarker for searching for biochemically similar TCRs in unlabeled
    datasets. 
    """ 

        # <cdr_col> usually 'cdr3_b_aa', we save this for later 
    cdr3_col = str(col)
        # <pw_mat> pairwise matrix, typically the network between antigen enriched clones
    pw_mat = getattr(tr, pw_mat_str)# defaults to 'pw_beta',
        # <rw_mat> rectangular matrix, typically the network betwen antigen enriched and bulk clones
    rw_mat = getattr(tr, rw_mat_str)# defaults to 'rw_beta',
        # remind the user if they are using sparse or dense matrices as input
    print(f"PW MATRIX TYPE = {type(pw_mat)}")
    print(f"RW MATRIX TYPE = {type(rw_mat)}")
        # <max_radius> int describing the highest distances, only will be needed for sparse impementation
        # <thresholds> array of thresholds radii to consider 
    thresholds = np.unique(thresholds)
    max_radius = thresholds.max() + 1
        # <weights> inverse probability weights, if none provided assume all are one
    if weights is None:
        warnings.warn("No weights provided trying to use tr_background.clone_df.weights")
        if 'weights' in tr_background.clone_df.columns:
            warnings.warn("USING tr_background.clone_df.weights")
            weights = tr_background.clone_df['weights']
        else:
            warnings.warn("SOMETHING IS PROBABLY WRONG!!!!. You don't have a 'weights' column in your tr_background clone_df, setting to 1, results could be biased if this data is V-J matched")
            weights = np.ones(rw_mat.shape[1])
    thresholds = np.unique(thresholds)
        # for each TCR, we calculate a empirical cummulative 
        # density function along a range of threshold radii



    # 88888b.   .d88b.  888  888  888 
    # 888 "88b d8P  Y8b 888  888  888 
    # 888  888 88888888 888  888  888 
    # 888  888 Y8b.     Y88b 888 d88P 
    # 888  888  "Y8888   "Y8888888P" 

    # KMB: I’ve confirmed that the distance_ecdf(absolute_weight = True) and my function provide same ecdfs. Therefore distance_ecdf can be our ONLY ecdf function (which is nice because it will be used for both plots and results).  My function now relies on it. I have commented out my code but will remove completely in a later version.
    
    from tcrdist.ecdf import distance_ecdf
    thresholds, ecdfs2 = distance_ecdf(pwrect =rw_mat, 
      thresholds = thresholds, 
      weights= weights, 
      pseudo_count=0, 
      skip_diag = False, 
      absolute_weight = True)
    
    ecdfs2 = [pd.Series(x, index = thresholds) for x in ecdfs2]
    max_radi2 = [x[x<=ctrl_bkgd].last_valid_index() for x in ecdfs2]
    # TO SOLVE NOTABLE BUG IN THE ABOVE LINE!! If a radius is None (the next line will fail, thus set Nones to 0.
    max_radi2 = [x if (x is not None) else 0 for x in max_radi2]

    if forced_max_radius is not None:
        max_radi2 = [min(x,forced_max_radius) for x in max_radi2]

    #          888      888 
    #          888      888 
    #          888      888 
    #  .d88b.  888  .d88888 
    # d88""88b 888 d88" 888 
    # 888  888 888 888  888 
    # Y88..88P 888 Y88b 888 
    #  "Y88P"  888  "Y88888 

    # print("USING : _compute_weighted_pop_estimate_ecdf_rowwise")
    # ecdfs = parmap.map(_compute_weighted_pop_estimate_ecdf_rowwise, range(0,rw_mat.shape[0]), 
    #    data = rw_mat, 
    #    weights = weights,
    #    thresholds = thresholds, 
    #    pm_pbar = True, 
    #    pm_processes = ncpus)
    # #<max_radi> Based on acceptable ctrl_bkgd, we find max acceptable radius from each TCR
    # max_radi = [x[x<=ctrl_bkgd].last_valid_index() for x in ecdfs]
    # # TO SOLVE NOTABLE BUG IN THE ABOVE LINE!! If a radius is None (the next line will fail, thus set Nones to 0.
    # max_radi = [x if (x is not None) else 0 for x in max_radi]  
    
    # if forced_max_radius is not None:
    #     max_radi = [min(x,forced_max_radius) for x in max_radi]
    # old vs. new methods must yield same results, 
    # TODO: remove old in next version, so there is a commit record
    #import pdb; pdb.set_trace()
    #assert np.all(max_radi2 == max_radi)
    #print("SUCCESS COMPARING METHODS")
    max_radi = max_radi2
        # <target_hits> number of hits within the antigen enriched repertoire at the control radius 
    if scipy.sparse.issparse(pw_mat):
        # sparse implementation, relies on _todense_row (see above in this module)
        target_hits    = [np.sum(_todense_row(pw_mat[i,:], max_radius )[0] <= t) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
    else: 
        target_hits    = [np.sum(pw_mat[i,:] <= t) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
        # <bkgd_hits>  number of hits within the bulk unenriched repertoire at the control radius 
        # <bkgd_hits_weighted> number of hits when inverse probability weights are applied
    if scipy.sparse.issparse(rw_mat):
        bkgd_hits          = [np.sum(_todense_row(rw_mat[i,:], max_radius)[0] <= t) for t,i in zip(max_radi, range(0,rw_mat.shape[0]))]
        bkgd_hits_weighted = [np.sum(1*(_todense_row(rw_mat[i,:], max_radius)[0] <= t) * (weights)) for t,i in zip(max_radi, range(0,rw_mat.shape[0]))]
    else:
        bkgd_hits      = [np.sum(rw_mat[i,:] <= t) for t,i in zip(max_radi, range(0,rw_mat.shape[0]))]
        bkgd_hits_weighted = [np.sum(1*(rw_mat[i,:] <= t) * (weights)) for t,i in zip(max_radi, range(0,rw_mat.shape[0]))]

        # <bkdg_total> number of rows in background
    bkgd_total = rw_mat.shape[1]
        # <bkgd_weighted_total > 
    bkgd_weighted_total = rw_mat.shape[1] # In modern, implemenation demoninator is simply length not np.sum(weights)
        # <ctrl> divide total hits by total posible hits
    ctrl = np.array(bkgd_hits) / bkgd_total 
        # <ctrl_weighted> divide weighted hits byt total possible hits to approximate expected frequency in a real repertoire
    ctrl_weighted= np.array(bkgd_hits_weighted)/bkgd_weighted_total 
        # <center_df> this is the kernal of dataframe
    
        # we typically want pgen in our analysis so
    if f'pgen_{col}' in tr.clone_df.columns:
        centers_df = pd.DataFrame({col: tr.clone_df[col].copy(),
                  add_cols[0] : tr.clone_df[add_cols[0]].copy(), 
                  add_cols[1] : tr.clone_df[add_cols[1]].copy(), 
                  'pgen' : tr.clone_df[f'pgen_{col}'],
                  'radius': max_radi,
                  'target_hits' : target_hits, 
                  'bkgd_hits': bkgd_hits,
                  'bkgd_hits_weighted': bkgd_hits_weighted,
                  'bkgd_total' :  bkgd_total,
                  'ctrl' : ctrl,
                  'ctrl_weighted' : ctrl_weighted})
    else: 
        print("YOU DID NOT PRECALCULATE OPTIONAL PGENS SEE auto_pgen(TCRrep) NEXT TIME")
        centers_df = pd.DataFrame({col: tr.clone_df[col].copy(),
                  add_cols[0] : tr.clone_df[add_cols[0]].copy(), 
                  add_cols[1] : tr.clone_df[add_cols[1]].copy(), 
                  'radius': max_radi,
                  'target_hits' : target_hits, 
                  'bkgd_hits': bkgd_hits,
                  'bkgd_hits_weighted': bkgd_hits_weighted,
                  'bkgd_total' :  bkgd_total,
                  'ctrl' : ctrl,
                  'ctrl_weighted' : ctrl_weighted})

    # Total bakground counts
    n1 = tr.clone_df.shape[0]   
    n2 = bkgd_weighted_total 
    print(f"ANTIGEN-ENRICHED CLONES : {n1}")
    print(f"BULK  BACKGROUND CLONES : {n2}")  
    print("COMPUTING WEIGHTED ODDS RATIO, RELATIVE RATE")
    centers_df['target_misses'] = centers_df['target_hits'].apply(lambda x : n1-x)
    centers_df['TR'] = [compute_rate(pos=r['target_hits'], neg=(n1-r['target_hits'])) for i,r in centers_df.iterrows()]
    centers_df['TR2'] = [compute_rate(pos=r['target_hits'], neg=r['target_misses']) for i,r in centers_df.iterrows()]
    centers_df['BR_weighted'] = [compute_rate(pos=r['bkgd_hits_weighted'], 
                                    neg=n2-r['bkgd_hits_weighted']) for i,r in centers_df.iterrows()]
    centers_df['RR_weighted'] = centers_df['TR']/centers_df['BR_weighted']
    centers_df['OR_weighted'] =[compute_odds_ratio(pos=r['target_hits'], 
                                       neg=n1-r['target_hits'], 
                                       bpos=r['bkgd_hits_weighted'], 
                                       bneg= n2-r['bkgd_hits_weighted'], ps = 1) for i,r in centers_df.iterrows()]
    print("COMPUTING CHI-SQUARE STATISTIC BASED ON WEIGHTED DISTANCE HITS <chi2dist>")
    centers_df['chi2dist'] = [chi2_contingency(np.array(
            [[1+r['target_hits'], 1+n1-r['target_hits']],[1+r['bkgd_hits_weighted'], 1+n2-r['bkgd_hits_weighted']]]))[0] for _,r in centers_df.iterrows() ]

    if include_seq_info:
        print("INCLUDING SEQ INFO")
        if scipy.sparse.issparse(pw_mat):
            target_neighbors = [list((_todense_row(pw_mat[i,:], max_radius)[0] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
        else:
            target_neighbors = [list((pw_mat[i,:] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
        
        centers_df['target_neighbors'] = target_neighbors 
        centers_df['target_seqs'] = [tr.clone_df.loc[ar,col].to_list() for ar in target_neighbors]
        if scipy.sparse.issparse(rw_mat):
            background_neighbors = [list((_todense_row(rw_mat[i,:], max_radius)[0] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
        else:
            background_neighbors = [list((rw_mat[i,:] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,pw_mat.shape[0]))]
        

        centers_df['background_neighbors'] = background_neighbors
        #col = "cdr3_b_aa"
        centers_df['background_seqs'] = [tr_background.clone_df.loc[ar,col].to_list() for ar in background_neighbors]
        vcol = add_cols[0] #v_b_gene 
        centers_df['background_v'] = [tr_background.clone_df.loc[ar,vcol].to_list() for ar in background_neighbors]
        jcol = add_cols[1] #j_b_gene
        centers_df['background_j'] = [tr_background.clone_df.loc[ar,jcol].to_list() for ar in background_neighbors]
        adjcol  = 'adj_freq_pVJ'
        if adjcol in tr_background.clone_df.columns:
            centers_df['adj_freq'] = [tr_background.clone_df.loc[ar,adjcol].to_list() for ar in background_neighbors]
    
    if generate_regex:
        print("GENERATING REGEX FOR EACH META-CLONOTYPE")
        #cdr3_col = "cdr3_b_aa"
        centers_df['regex'] = [_index_to_regex_str(ind = r['target_neighbors'], 
                clone_df = tr.clone_df, 
                pwmat = None, 
                col = cdr3_col, 
                centroid = tr.clone_df[cdr3_col][i],
                ntrim = 3,
                ctrim = 2,
                max_ambiguity = 5) for i,r in centers_df.iterrows()]

    if test_regex:
        print("TESTING REGEX FOR EACH META-CLONOTYPE")
        # THIS INVOLVES TESTING REGEX AGAINST TARGET AND BACKGROUND
        # Test regex against backgound
        rs = centers_df['regex'].to_list()
        bkgd_cdr3 = tr_background.clone_df[cdr3_col].to_list()
        target_cdr3 = tr.clone_df[cdr3_col].to_list()
        target_regex_hits = parmap.map(_multi_regex, rs, bkgd_cdr3 = target_cdr3 , pm_pbar = True, pm_processes = ncpus)
        target_re_hits = [np.sum([1 if (x is not None) else 0 for x in l]) for l in target_regex_hits ]
        centers_df['target_re_hits'] = target_re_hits
        bkgd_regex_hits = parmap.map(_multi_regex, rs, bkgd_cdr3 = bkgd_cdr3, pm_pbar = True, pm_processes = ncpus)
        bkgd_re_hits = [np.sum([1 if (x is not None) else 0 for x in l]) for l in bkgd_regex_hits]
        centers_df['bkgd_re_hits'] = bkgd_re_hits
        bkgd_weighted_re_hits = [np.sum(np.array([1 if (x is not None) else 0 for x in l]) * (weights)) for l in bkgd_regex_hits] 
        # Compute Relative Rates
        print("COMPUTING WEIGHTED ODDS RATIO, RELATIVE RATES FOR EACH REGEX")
        centers_df['bkgd_re_weighted_hits'] = bkgd_weighted_re_hits 
        centers_df['TR_re'] = [compute_rate(pos=r['target_re_hits'], neg=n1-r['target_re_hits']) for i,r in centers_df.iterrows()]
        centers_df['BR_re_weighted'] = [compute_rate(pos=r['bkgd_re_weighted_hits'], 
                                     neg=n2-r['bkgd_re_weighted_hits']) for i,r in centers_df.iterrows()]
        centers_df['RR_re_weighted'] = centers_df['TR']/centers_df['BR_re_weighted']
        centers_df['OR_re_weighted'] =[compute_odds_ratio(pos=r['target_re_hits'], 
                                       neg=n1-r['target_re_hits'], 
                                       bpos=r['bkgd_re_weighted_hits'], 
                                       bneg= n2-r['bkgd_re_weighted_hits'], ps = 1) for i,r in centers_df.iterrows()]
        print("COMPUTING CHI-SQUARE STATISTIC BASED ON WEIGHTED REGEX HITS <chi2re>")
        centers_df['chi2re'] = [chi2_contingency(np.array(
            [[1+r['target_re_hits'], 1+n1-r['target_re_hits']],[1+r['bkgd_re_weighted_hits'], 1+n2-r['bkgd_re_weighted_hits']]]))[0] for _,r in centers_df.iterrows() ]
        print(f"COMPUTING JOINT CHI-SQUARE STATISTIC <chi2joint>BASED ON WEIGHTED REGEX HITS {beta_re} * <chi2re> + {beta_dist} * <chi2dist")
        centers_df['chi2joint'] = [beta_re  * r['chi2re'] + beta_dist* r['chi2dist'] for _,r in centers_df.iterrows() ]



    return centers_df #, thresholds, ecdfs











#          888      888                       888                                            d8b          888                 888 
#          888      888                       888                                            Y8P          888                 888 
#          888      888                       888                                                         888                 888 
#  .d88b.  888  .d88888                   .d88888  .d88b.  88888b.  888d888 .d88b.   .d8888b 888  8888b.  888888 .d88b.   .d88888 
# d88""88b 888 d88" 888                  d88" 888 d8P  Y8b 888 "88b 888P"  d8P  Y8b d88P"    888     "88b 888   d8P  Y8b d88" 888 
# 888  888 888 888  888      888888      888  888 88888888 888  888 888    88888888 888      888 .d888888 888   88888888 888  888 
# Y88..88P 888 Y88b 888                  Y88b 888 Y8b.     888 d88P 888    Y8b.     Y88b.    888 888  888 Y88b. Y8b.     Y88b 888 
#  "Y88P"  888  "Y88888                   "Y88888  "Y8888  88888P"  888     "Y8888   "Y8888P 888 "Y888888  "Y888 "Y8888   "Y88888 
#                                                          888                                                                    
#                                                          888                                                                    
#                                                          888                                                        
# 
# THIS WAS ONCE USED FOR COMPUTING ECDFS BUT WE CAN NOW RELY ONLY ON distance_ecdf

# def compute_population_estimate_ecdf(data, weights = None, thresholds=None):
#     """
#     """
#     if thresholds is None:
#         thresholds = np.unique(data[:])
  
#     if weights is None:
#         weights = np.ones(data.shape[0])
#     else:
#         weights = np.array(weights)
    
#     """Vectorized and faster, using broadcasting for the <= expression"""
#     # Step 1 : compare a set of n distances to a set of say 25 threshold 
#         # (n dists x 1) broadcast against (1 x 25 thresholds) ==> (n, 25), where each column representd number of hits below the threshold 
#     M= 1*(data[:, None] <= thresholds[None, :])
#     # Step 2 : Adjust hit be relative weight (i.e. actual_freq_vj / sampled_freq_vj)
#         # Broadcast (n, 25) on weights (n, 1) ==> (n, 25)
#     M = M * weights[:,None]
#     # Step 3: Take row sums, producing the weighted sum at each threshold ==> (25,)
#     M =  np.sum(M, axis = 0)
#     # Step 4: Divide by the weighted total (i.e., the amount if all seqs were present, which because of enrichment is far greater than 1)
#     # print(M)   
#     ecdf = M / len(weights) 
#     return ecdf


# def _compute_weighted_pop_estimate_ecdf_rowwise(i, thresholds, weights, data, max_radius = 50):
#     """
#     This is a wrapper for compute_ecdf using parmap
#     """
#     if scipy.sparse.issparse(data):
#         row = data[i,:].toarray()[0]
#         row[row==0] = (max_radius + 1)
#         row[row==-1] = 0
#         ec = compute_population_estimate_ecdf(data = row,  weights = weights, thresholds = thresholds)
#     else:
#         ec = compute_population_estimate_ecdf(data = data[i,:],  weights = weights, thresholds = thresholds)
#     return pd.Series(ec, index = thresholds)

# def _compute_pop_estimate_ecdf_rowwise(i, thresholds, data, max_radius = 50):
#     """
#     This is a warpper for compute_ecdf using parmap
#     """
#     if scipy.sparse.issparse(data):
#         row = data[i,:].toarray()[0]
#         row[row==0] = (max_radius + 1)
#         row[row==-1] = 0
#         ec = compute_population_estimate_ecdf(data = row, thresholds = thresholds)
#     else:
#         ec = compute_population_estimate_ecdf(data = data[i,:], thresholds = thresholds)
#     return pd.Series(ec, index = thresholds)


              
#                   888                                           888                          888      888 
#                   888                                           888                          888      888 
#                   888                                           888                          888      888 
#  .d88b.  888  888 888888 888d888 .d88b.  88888b.d88b.   .d88b.  888 888  888         .d88b.  888  .d88888 
# d8P  Y8b `Y8bd8P' 888    888P"  d8P  Y8b 888 "888 "88b d8P  Y8b 888 888  888        d88""88b 888 d88" 888 
# 88888888   X88K   888    888    88888888 888  888  888 88888888 888 888  888 888888 888  888 888 888  888 
# Y8b.     .d8""8b. Y88b.  888    Y8b.     888  888  888 Y8b.     888 Y88b 888        Y88..88P 888 Y88b 888 
#  "Y8888  888  888  "Y888 888     "Y8888  888  888  888  "Y8888  888  "Y88888         "Y88P"  888  "Y88888 
#                                                                          888                              
#                                                                     Y8b d88P                              
#                                                                      "Y88P"                               

# ## SOON TO BE DEPRECIATED BELOW THIS POINT: DO NOT USE

# def compute_ecdf(data, weights = None, counts=None, thresholds=None):
#     """
#     Computes the empirical cumulative distribution function at pre-specified
#     thresholds. Assumes thresholds is sorted and should be unique.

#     data : np.array

#     threshold : np.array or None
    
#     counts : np.array or None

#     weights :  np.array or None
#          vector of W_vj  (i.e  actual_freq_vj / sampled_freq_vj)
    
#     Example
#     -------
#     >>> compute_ecdf(data = np.array([1,2,3,4,5,6]), counts =np.array([100,100,100,100,100,100]), thresholds = np.array([3]))
#     array([0.5])
#     >>> compute_ecdf(data = np.array([1,2,3,4,5,6]), weights =np.array([1,1,1,1,1,1]), counts =np.array([100,100,100,100,100,100]), thresholds = np.array([3]))
#     array([0.5])
#     >>> compute_ecdf(data = np.array([1,2,3,4,5,6]), weights =np.array([10,1,1,1,1,1]), counts =np.array([100,100,100,100,100,100]), thresholds = np.array([3]))
#     array([0.8])

#     Notes
#     -----
    
#     Weights are used when we enriche for V,J combinations that are rare. 
#     For instance if some V,J pairing 
#     occurs at an actual frequency of 1/100 but we provide it at 1/10, 
#     we provide an enrichment weight 
#     (actual_freq_vj / sampled_freq_v)


#     -----
#     """

#     if thresholds is None:
#         thresholds = np.unique(data[:])
#     if counts is None:
#         counts = np.ones(data.shape)
    
#     if weights is None:
#         weights = np.ones(data.shape)
#     else:
#         weights = np.array(weights)
    
#     tot = np.sum(counts * weights) 
#     """Vectorized and faster, using broadcasting for the <= expression"""
#     # Step 1 : compare a set of n distances to a set of say 25 threshold 
#         # (n dists x 1) broadcast against (1 x 25 thresholds) ==> (n, 25), where each column representd number of hits below the threshold 
#     M= 1*(data[:, None] <= thresholds[None, :])
#     # Step 2 : Adjust hit be relative weight (i.e. actual_freq_vj / sampled_freq_vj)
#         # Broadcast (n, 25) on weights (n, 1) ==> (n, 25)
#     M = M * weights[:,None] * counts[:,None]
#     # Step 3: Take row sums, producing the weighted sum at each threshold ==> (25,)
#     M =  np.sum(M, axis = 0)
#     # Step 4: Divide by the weighted total (i.e., the amount if all seqs were present, which because of enrichment is far greater than 1)
#     ecdf = M / tot

#     return ecdf



# def _compute_ecdf_rowwise(i, thresholds, data):
#     """
#     This is a warpper for compute_ecdf using parmap
#     """
#     ec = compute_ecdf(data = data[i,:], thresholds = thresholds)
#     return pd.Series(ec, index = thresholds)

# def _compute_weighted_ecdf_rowwise(i, thresholds, weights, data):
#     """
#     This is a warpper for compute_ecdf using parmap
#     """
#     ec = compute_ecdf(data = data[i,:], weights = weights, thresholds = thresholds)
#     return pd.Series(ec, index = thresholds)







# def bkgd_cntl_nn(  tr, 
#                    tr_background,
#                    weights = None,
#                    ctrl_bkgd = 2*10**-5, 
#                    col = 'cdr3_b_aa',
#                    add_cols = ['v_b_gene', 'j_b_gene'],
#                    ncpus = 4,
#                    thresholds = [x for x in range(0,50,2)],
#                    generate_regex = True,
#                    test_regex = True):   

#     """
#     bkgd_cntl_nn standfs for background controlled nearest neighbors. 

#     tr : tcrdist.repertoire.TCRrep
#         TCRrep instance with clone_df of target data
#     tr_background : tcrdist.repertoire.TCRrep
#         TCRrep instance with clone_df of bulk data
#     ctrl_bkgd : float
#         Default is 2*10**-5, the acceptable level of background neighbror 
#     col : str
#         Default is cdr3_b_aa', the column containing CDR3 string
#     add_cols  : list
#         Extra columns from clone_df to include in centers_df['v_b_gene', 'j_b_gene'],
#     ncpus = 4
#         passed to pm_processes when using parmap   
#     thresholds : list
#         Default is [x for x in range(0,50,2)] indicating tcrdist thresholds to compute 
#         a ecdf over.
#     generate_regex : bool
#         if True, generate a regex pattern capturing most if not all of 
#         sequences in the neigbhorhood
#     test_regex = True
#         if True, test regex against both CDR3 in target and background

#     Notes
#     -----

#     The logic behind this function is that we want to find radi that 
#     are appropriate for each TCR in a epitope-specific target set 
#     that control the frequency of finding neighbors in a background set. 
#     For instance, a TCR with common V gene usage and a high probabililty 
#     of generation CDR3, will potentially have many nieghbors in a 
#     pre-selection background set, whereas a TCR with rare gene usage 
#     and a low probability of generation CDR3 can acomodate a larger 
#     neighborhood radi. 

#     The result dataframe can be sorted to find TCRs that anchor 
#     a nieghborhood of TCRs including the most epitope specific
#     nieghbors. We hypothesis that these TCRs are more likely to serve as a useful 
#     biomarker for searching for biochemically similar TCRs in unlabeled
#     datasets. 
#     """ 
#     thresholds = np.unique(thresholds)
#     # for each TCR, we calculate a empirical cummulative 
#     # density function along a range of threshold radi
#     if weights is None:
#         ecdfs = parmap.map(_compute_ecdf_rowwise, range(0,tr.rw_beta.shape[0]), 
#             data = tr.rw_beta, 
#             thresholds = thresholds, 
#             pm_pbar = True, 
#             pm_processes = ncpus)
#     else: 
#         ecdfs = parmap.map(_compute_weighted_ecdf_rowwise, range(0,tr.rw_beta.shape[0]), 
#             data = tr.rw_beta, 
#             weights = weights,
#             thresholds = thresholds, 
#             pm_pbar = True, 
#             pm_processes = ncpus)

#     # Based on acceptable ctrl_bkgd, we find max acceptable radi from each TCR
#     max_radi = [x[x<=ctrl_bkgd].last_valid_index() for x in ecdfs]
#     # THERE IS A NOTABLE BUG IN THE ABOVE LINE!! IF a radius is None (the next line will fail, thus set Nones to 0.
#     max_radi = [x if (x is not None) else 0 for x in max_radi]

#     # Confirm that this procedure did infact control background hits
#     # NOTE WE CAN NO LONGER GAURANTEE THIS IS TRUE IF NUMBER OF PERFECT MATCHES EXCEEDS 0
#     # assert np.all([ np.sum(tr.rw_beta[i,:] <= t) < tr.rw_beta.shape[1]*ctrl_bkgd for t,i in zip(max_radi, range(0,tr.rw_beta.shape[0]))])
#     # Tabulate target-set hits. (That is hits within epitope specific set)  
    
    
#     if weights is None:
#         target_hits = [np.sum(tr.pw_beta[i,:] <= t) for t,i in zip(max_radi, range(0,tr.pw_beta.shape[0]))]
#         background_hits = [np.sum(tr.rw_beta[i,:] <= t) for t,i in zip(max_radi, range(0,tr.rw_beta.shape[0]))]
#     else:
#         target_hits              = [np.sum(tr.pw_beta[i,:] <= t) for t,i in zip(max_radi, range(0,tr.pw_beta.shape[0]))]
#         target_weighted_hits     = [np.sum(tr.pw_beta[i,:] <= t) for t,i in zip(max_radi, range(0,tr.pw_beta.shape[0]))]
#         background_hits          = [np.sum(tr.rw_beta[i,:] <= t) for t,i in zip(max_radi, range(0,tr.rw_beta.shape[0]))]
#         background_weighted_hits = [np.sum(1*(tr.rw_beta[i,:] <= t) * (1/weights)) for t,i in zip(max_radi, range(0,tr.rw_beta.shape[0]))]
    
#     target_neighbors =   [list((tr.pw_beta[i,:] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,tr.pw_beta.shape[0]))]
#     target_seqs = [tr.clone_df.loc[ar,col].to_list() for ar in target_neighbors]
#     background_neighbors = [list((tr.rw_beta[i,:] <= t).nonzero()[0]) for t,i in zip(max_radi, range(0,tr.pw_beta.shape[0]))]
#     background_seqs = [tr_background.clone_df.loc[ar,col].to_list() for ar in background_neighbors]

    

#     centers_df = pd.DataFrame({col: tr.clone_df[col].copy(),
#                   add_cols[0] : tr.clone_df[add_cols[0]].copy(), 
#                   add_cols[1] : tr.clone_df[add_cols[1]].copy(), 
#                   'pgen' : tr.clone_df[f'pgen_{col}'],
#                   'bkgd_hits': background_hits,
#                   'background_hits': background_hits, # redundant so we don't break anything
#                   'target_hits' : target_hits, 
#                   'max_radi': max_radi,
#                   'target_neighbors': target_neighbors,
#                   'target_seqs': target_seqs,
#                   'background_neighbors': background_seqs})
    
#     if weights is not None:
#         centers_df['target_weighted_hits'] = target_weighted_hits
#         centers_df['bkgd_weighted_hits'] = background_weighted_hits


#     if generate_regex:
       
#         centers_df['regex'] = [_index_to_regex_str(ind = r['target_neighbors'], 
#                 clone_df = tr.clone_df, 
#                 pwmat = None, 
#                 col = col, 
#                 centroid = tr.clone_df[col][i],
#                 ntrim = 3,
#                 ctrim = 2,
#                 max_ambiguity = 5) for i,r in centers_df.iterrows()]

#     if test_regex:
#         from tcrdist.regex import _index_to_regex_str, _index_to_seqs 
#         # Test regex against backgound
#         rs = centers_df['regex'].to_list()
#         bkgd_cdr3 = tr_background.clone_df[col].to_list()
        
#         target_cdr3 = tr.clone_df[col].to_list()
#         target_regex_hits = parmap.map(_multi_regex, rs, bkgd_cdr3 = target_cdr3 , pm_pbar = True, pm_processes = ncpus)
#         target_re_hits = [np.sum([1 if (x is not None) else 0 for x in l]) for l in target_regex_hits ]
#         centers_df['target_re_hits'] = target_re_hits

#         bkgd_regex_hits = parmap.map(_multi_regex, rs, bkgd_cdr3 = bkgd_cdr3, pm_pbar = True, pm_processes = ncpus)
#         bkgd_re_hits = [np.sum([1 if (x is not None) else 0 for x in l]) for l in bkgd_regex_hits]
#         centers_df['bkgd_re_hits'] = bkgd_re_hits

#         if weights is not None:       
#             #target_weighted_re_hits = [np.array([1 if (x is not None) else 0 for x in l]) * (1/weights) for l in target_regex_hits] #1*[np.sum(1*(tr.rw_beta[i,:] <= t) * (1/weights)) for t,i in zip(max_radi, range(0,tr.rw_beta.shape[0]))]
#             #centers_df['target_weighted_re_hits'] = target_weighted_re_hits
#             bkgd_weighted_re_hits = [np.sum(np.array([1 if (x is not None) else 0 for x in l]) * (1/weights)) for l in bkgd_regex_hits] 
#             centers_df['bkgd_weighted_re_hits'] = bkgd_weighted_re_hits 

#     n1 = tr.clone_df.shape[0]
#     # Total bakground is just the number of clones on background
#     n2 = tr_background.clone_df.shape[0]
#     centers_df['TR'] =  [compute_rate(pos=r['target_hits'], neg=n1-r['target_hits']) for i,r in centers_df.iterrows()]
#     centers_df['BR'] = [compute_rate(pos=r['bkgd_hits'], neg= n2-r['bkgd_hits']) for i,r in centers_df.iterrows()]
#     centers_df['RR'] = centers_df['TR']/centers_df['BR']
#     centers_df['OR'] = [compute_odds_ratio(pos=r['target_hits'], 
#                                          neg=n1-r['target_hits'], 
#                                          bpos=r['bkgd_hits'], 
#                                          bneg= n2-r['bkgd_hits'], ps = 1) for i,r in centers_df.iterrows()]
#     centers_df['ctrl'] = centers_df['bkgd_hits'] / n2

#     if weights is not None:
#         # Total bakground sum of the inverse of weightes
#         n1 = tr.clone_df.shape[0]
#         n2 = np.sum(1/weights)
#         centers_df['TR_weighted'] = [compute_rate(pos=r['target_weighted_hits'], neg=n1-r['target_weighted_hits']) for i,r in centers_df.iterrows()]
#         centers_df['BR_weighted'] = [compute_rate(pos=r['bkgd_weighted_hits'], neg=n2-r['bkgd_weighted_hits']) for i,r in centers_df.iterrows()]
#         centers_df['RR_weighted'] = centers_df['TR_weighted']/centers_df['BR_weighted']
#         centers_df['OR_weighted'] =[compute_odds_ratio(pos=r['target_hits'], 
#                                                neg=n1-r['target_hits'], 
#                                                bpos=r['bkgd_weighted_hits'], 
#                                                bneg= n2-r['bkgd_weighted_hits'], ps = 1) for i,r in centers_df.iterrows()]
#         centers_df['ctrl_weighted'] = centers_df['bkgd_weighted_hits'] / n2


#     return centers_df



import pytest

def test_example_10_sparse():
    """
    If you just want a 'tcrdistances' of some target seqs against another set computed sparsely

    (1) cell_df is asigned the first 10 cells in dash.csv
    (2) compute tcrdistances with default settings.
    (3) compute rectangular distance between clone_df and df2.
    (4) compute rectangular distance between clone_df and any 
    arbtirary df3, which need not be associated with the TCRrep object.
    (5) compute rectangular distance with only a subset of the TCRrep.clone_df
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import pw2dense
    import numpy as np

    df = pd.read_csv("dash.csv")
    df2 = pd.read_csv("dash2.csv")
    df = df.head(10)                        #(1)
    tr = TCRrep(cell_df = df,               #(2)
                df2 = df2, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    tr.compute_rect_distances(df = tr.clone_df, df2 = df2)
    assert tr.rw_alpha.shape == (10,1924) 
    assert tr.rw_beta.shape  == (10,1924)

    rw_alpha = tr.rw_alpha.copy()
    rw_beta = tr.rw_beta.copy()

    radius = 100
    tr.cpus = 1
    tr.compute_sparse_rect_distances(df = tr.clone_df, df2 = df2, radius = radius)
    d = pw2dense(tr.rw_alpha, radius)
    print(rw_alpha[:2, :10])
    print(tr.rw_alpha.todense()[:2, :10])
    print(d[:2, :10])
    assert np.all(rw_alpha[rw_alpha <= radius] == d[d <= radius])

    d = pw2dense(tr.rw_beta, radius)
    assert np.all(rw_beta[rw_beta <= radius] == d[d <= radius])

    
    radius = 5000
    tr.compute_sparse_rect_distances(df = tr.clone_df, df2 = df2, radius = radius)
    d = pw2dense(tr.rw_alpha, radius)
    print(rw_alpha[:2, :10])
    print(tr.rw_alpha.todense()[:2, :10])
    assert np.all(rw_alpha == d)

    d = pw2dense(tr.rw_beta, radius)
    assert np.all(rw_beta == d)
    

def test_example_10_sparse_multiprocessing():
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import pw2dense
    import numpy as np

    df2 = pd.read_csv("dash2.csv")
    tr = TCRrep(cell_df = df2,               #(2)
                df2 = df2, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    tr.compute_rect_distances(df = tr.clone_df, df2 = df2)
    assert tr.rw_alpha.shape == (1924,1924) 
    assert tr.rw_beta.shape  == (1924,1924)

    rw_alpha = tr.rw_alpha.copy()
    rw_beta = tr.rw_beta.copy()

    radius = 150
    tr.cpus = 2
    tr.compute_sparse_rect_distances(df = tr.clone_df, df2 = df2, radius = radius)
    d = pw2dense(tr.rw_alpha, radius)
    assert np.all(rw_alpha[rw_alpha <= radius] == d[d <= radius])

    d = pw2dense(tr.rw_beta, radius)
    assert np.all(rw_beta[rw_beta <= radius] == d[d <= radius])


import pytest 



def test_downloads():
    from tcrdist.setup_tests import list_available_zip_files  
    list_available_zip_files()
    """
    ['dash.zip',
    'human_T_alpha_beta_sim200K.zip',
    'vdjDB_PMID28636592.zip',
    'sant.csv.zip',
    'bulk.csv.zip',
    'wiraninha_sampler.zip',
    'ruggiero_mouse_sampler.zip',
    'ruggiero_human_sampler.zip',
    'britanova_human_beta_t_cb.tsv.sampler.tsv.zip',
    'emerson_human_beta_t_cmvneg.tsv.sampler.tsv.zip',
    'ruggiero_human_alpha_t.tsv.sampler.tsv.zip',
    'ruggiero_human_beta_t.tsv.sampler.tsv.zip']
    """
    from tcrdist.setup_tests import download_and_extract_zip_file
    """
    Get dash.zip contents
    """
    download_and_extract_zip_file( 'dash.zip', source = "dropbox", dest = ".")
    """
    Assert that file has been downloaded
    """
    import os
    assert os.path.isfile("dash.zip")
    """
    Assert that individual files of successively been inflated
    """
    assert os.path.isfile("dash.csv")
    assert os.path.isfile("dash2.csv")
    assert os.path.isfile("dash_human.csv")
    assert os.path.isfile("dash_beta_airr.csv")






def test_example_tree_args():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(100).reset_index(drop = True)

    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.mouse.ab.tree.html')

    tcrtree.build_tree()

    assert os.path.isfile('dash.mouse.ab.tree.html')


import pytest 



def test_example_2():
    """  
    If you want 'tcrdistances' with changes over some parameters.
    For instance you want to change the gap penalty on CDR3s to 5. 
    """
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    
    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                compute_distances = False,
                db_file = 'alphabeta_gammadelta_db.tsv')

    tr.kargs_a['cdr3_a_aa']['gap_penalty'] = 5 
    tr.kargs_b['cdr3_b_aa']['gap_penalty'] = 5 
    
    tr.compute_distances()
    
    tr.pw_alpha
    tr.pw_beta


import pytest 



def test_workflow_2():
    """
    Load all the TCRs associated with a particular epitope in 
    the Adaptive Biotechnology COVID19 Data Release 2
    """
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.adpt_funcs import get_basic_centroids
    from tcrdist.paths import path_to_base
    path = os.path.join(path_to_base, 'tcrdist', 'data', 'covid19')
    file = 'mira_epitope_16_1683_QYIKWPWYI_YEQYIKWPW_YEQYIKWPWY.tcrdist3.csv'
    filename = os.path.join(path,file)
    
    df = pd.read_csv(filename, sep = ",")
   
    df = df[['cell_type','subject','v_b_gene','j_b_gene','cdr3_b_aa',
            'epitope', 'age', 'sex','cohort']]
            
    df['count'] = 1
    
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['beta'])
 
    tr = get_basic_centroids(tr, max_dist = 200)
    
    tr.centroids_df
    
    tr.clone_df['covid'] = ['healthy' if x.find("Healthy") != -1 else "covid" for x in tr.clone_df.cohort]
    
    from tcrdist.rep_diff import neighborhood_diff, hcluster_diff, member_summ
    import hierdiff
    #tr.clone_df['covid'] = ['healthy' if x.find("Healthy") != -1 else "covid" for x in tr.clone_df.cohort]
    #nd = neighborhood_diff(tr.clone_df, tr.pw_beta, x_cols = ['covid'], count_col = 'count')

    tr.clone_df['covid'] = ['healthy' if x.find("Healthy") != -1 else "covid" for x in tr.clone_df.cohort]
    res, Z= hcluster_diff(tr.clone_df, tr.pw_beta, x_cols = ['covid'], count_col = 'count')
    
    res_summary = member_summ(res_df = res, clone_df = tr.clone_df, addl_cols=['cohort','subject'])
    
    res_detailed = pd.concat([res, res_summary], axis = 1)
    
    html = hierdiff.plot_hclust_props(Z,
                title='PA Epitope Example',
                res=res_detailed,
                tooltip_cols=['cdr3_b_aa','v_b_gene', 'j_b_gene','cohort','subject'],
                alpha=0.05, 
                alpha_col='pvalue')
    
    with open('hierdiff_example.html', 'w') as fh:
        fh.write(html)



"""
test nieghbors
"""
import pytest

@pytest.mark.skip(reason="This test documents an example, but is redundant. Skipped in the interest of CI time.")
def test_current_example():
	import os
	import pandas as pd
	import numpy as np 
	from tcrdist.repertoire import TCRrep
	from tcrdist.neighbors import compute_ecdf, bkgd_cntl_nn2
	from tcrdist.automate import auto_pgen
	import scipy.sparse
	
	
	fn_mira_background  =  'mira_epitope_60_436_MWSFNPETNI_SFNPETNIL_SMWSFNPET.tcrdist3.csv.olga100K_brit100K_bkgd.csv'
	fn_mira_background  =   os.path.join('tcrdist', 'data', 'covid19', fn_mira_background)
	df_background       =   pd.read_csv(fn_mira_background)
	tr_background       =   TCRrep( cell_df = df_background.copy(), 
									organism = "human", 
									chains= ['beta'], 
									compute_distances = False)

	fn_mira             =  'mira_epitope_60_436_MWSFNPETNI_SFNPETNIL_SMWSFNPET.tcrdist3.csv'
	fn_mira             =   os.path.join('tcrdist', 'data', 'covid19', fn_mira)
	df_mira             =   pd.read_csv(fn_mira)
	df_mira             =   df_mira[['v_b_gene', 'j_b_gene', 'cdr3_b_aa']]
	tr                  =   TCRrep(	cell_df = df_mira.copy(), 
									organism = 'human', 
									chains = ['beta'], 
									db_file = 'alphabeta_gammadelta_db.tsv',
									store_all_cdr = False,
									compute_distances = True)
	auto_pgen(tr)

	tr.compute_rect_distances(df = tr.clone_df, 
							  df2 = tr_background.clone_df, 
							  store = False)

	assert tr.rw_beta.shape[0] == tr.clone_df.shape[0]

	centers_df = bkgd_cntl_nn2(	tr = tr, 
								tr_background = tr_background,
								ctrl_bkgd = 10**-6, 
								weights =tr_background.clone_df.weights,
								col = 'cdr3_b_aa',
								ncpus = 2,
								thresholds = [x for x in range(0,50,2)], 
								generate_regex = True, 
								test_regex = True)
	
	out_fn_center_df		     = 'mira_epitope_60_436_MWSFNPETNI_SFNPETNIL_SMWSFNPET.tcrdist3.csv.centers_df.csv' 
	out_fn_rw_beta_sparse_matrix = 'mira_epitope_60_436_MWSFNPETNI_SFNPETNIL_SMWSFNPET.tcrdist3.csv.rw_beta.sparse.npz' 
	centers_df.to_csv(out_fn_center_df, index = False)
	tr.rw_beta[tr.rw_beta == 0]  = 1 # set true zeros to 1
	tr.rw_beta[tr.rw_beta > 50]  = 0 # ignores everything less than 100
	rw_beta_sparse = scipy.sparse.csr_matrix(tr.rw_beta)
	scipy.sparse.save_npz(out_fn_rw_beta_sparse_matrix, rw_beta_sparse)

# NOTE USED ANYMORE
# def test_compute_population_estimate_ecdf():
# 	"""
# 	Test import, and ecdf with some or all optional arguments
# 	"""
# 	from tcrdist.neighbors import compute_population_estimate_ecdf
# 	import numpy as np
# 	assert isinstance(compute_population_estimate_ecdf(data = np.array([1,2,3,4,5,6])), np.ndarray)
# 	assert isinstance(compute_population_estimate_ecdf(data = np.array([1,2,3,4,5,6]), thresholds = np.array([3])), np.ndarray)
# 	r = compute_population_estimate_ecdf(data = np.array([1,2,3,4,5,6]), thresholds = np.array([3]))
# 	assert np.all( np.isclose(r, np.array([0.5]) ) )

@pytest.mark.skip(reason="This test documents an example, but is redundant. Skipped in the interest of CI time.")
def test_old_example():
	"""
	The purpose of this example is to show the use of 
	chosing thresholds based on background discovery rate
	"""
	import os
	import pandas as pd
	import numpy as np 
	from tcrdist.repertoire import TCRrep
	from tcrdist.neighbors import compute_ecdf, bkgd_cntl_nn2
	from tcrdist.automate import auto_pgen
	from tcrdist.regex import _index_to_regex_str, _index_to_seqs 
	from tcrdist.summarize import _summ, _dist_summ, _select, filter_gt, filter_is, test_for_subsets, test_for_almost_subsets

	fn = os.path.join('tcrdist', 'data', 'covid19', "m60_bkgd_test_input.csv")
	df_background = pd.read_csv(fn)
	
	tr_background = TCRrep(cell_df = df_background, 
					organism = "human", 
					chains= ['beta'], 
					compute_distances = False)
	tr_background.clone_df['weights'] = 1
	fn = os.path.join('tcrdist', 'data', 'covid19', "m60_test_input.csv")
	df = pd.read_csv(fn)
	
	tr = TCRrep(cell_df = df, 
	            organism = 'human', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv')
	
	auto_pgen(tr)

	tr.compute_rect_distances(df = tr.clone_df, 
							  df2 = tr_background.clone_df, 
							  store = False)

	assert tr.rw_beta.shape[0] == tr.clone_df.shape[0]

	centers_df = bkgd_cntl_nn2(	tr = tr, 
								tr_background = tr_background,
								ctrl_bkgd = 2*10**-5, 
								weights =tr_background.clone_df.weights,
								col = 'cdr3_b_aa',
								ncpus = 2,
								thresholds = [x for x in range(0,50,2)], 
								generate_regex = True, 
								test_regex = True)
 

	
	centers_df.sort_values(['target_hits'], ascending = False)
	
	# # Let's add some useful summary info about publicity and number of unique seqs
	# def tabulate_subjects(neighbor_df, clone_df, col_nn ='target_neighbors', col_seq= 'target_seqs'):
	# 	# Tabulate the number of unique subjects at each node
	# 	neighbor_df['nsubject'] = neighbor_df[col_nn].apply( lambda x: len(set(_select(clone_df, iloc_rows =x, col = 'subject'))))
	# 	# Tabulate the number of unique sequences at each node
	# 	neighbor_df['useq']    = neighbor_df[col_seq].apply( lambda x: len(set(x)))
	# 	return neighbor_df

	# tabulate_subjects(centers_df, tr.clone_df)
	
	# # Let's add some useful info about Probability of generation
	# centers_df['pgen_cdr3_b_aa'] = tr.clone_df.pgen_cdr3_b_aa.copy()
	# centers_df['pgen_dist'] = _summ(df = tr.clone_df, 
	# 	 indices = centers_df['target_neighbors'], 
	# 	 column = 'pgen_cdr3_b_aa', 
	# 	 f=_dist_summ)

	# from scipy.stats import chi2_contingency
	# n1 = 436
	# n2 = 100000
	# beta_re = 1 
	# beta_dist = 1 
	# centers_df['chi2re'] = [chi2_contingency(np.array( [[r['target_re_hits'], n1-r['target_re_hits']],[r['bkgd_re_hits'], n2-r['bkgd_re_hits']]]))[0] for _,r in centers_df.iterrows() ]
	# centers_df['chi2dist'] = [chi2_contingency(np.array( [[r['target_hits'], n1-r['target_hits']],[r['background_hits'], n2-r['background_hits']]]))[0] for _,r in centers_df.iterrows() ]
	# centers_df['chi2joint'] = [beta_re  * r['chi2re'] + beta_dist* r['chi2dist'] for _,r in centers_df.iterrows() ]

	# # Rank and select non-redundant
	# sorted_centers_df = centers_df.sort_values(['chi2joint'], ascending = False).copy()
	# sorted_centers_df['novel'] = test_for_almost_subsets(sorted_centers_df['target_neighbors'], 5)
	# sorted_filtered_centers_df = filter_is(sorted_centers_df, 'novel', 1).copy()
	# #sorted_filtered_centers_df.to_csv("m60_centers_df.csv")
	






import pytest



def longtest_pgen_with_parmap():
    """
    Test speed up of computation of many pgens using parmap to make use of more than one cpu

    For 1000 CDR3 

    Finished 'olga_in_series' in 26.3842 secs with 1 core
    Finished 'olga_in_parmap' in 6.2384 secs with 6 cores
    """
    import numpy as np
    import pandas as pd
    import parmap
    from tcrdist.pgen import OlgaModel
    from tcrdist.speed import timer
    from tcrdist.adpt_funcs import _valid_cdr3

    from tcrdist.setup_tests import download_and_extract_zip_file
    download_and_extract_zip_file( 'cdr3_beta_500K.zip', source = "dropbox", dest = ".")

    olga_beta  = OlgaModel(chain_folder = "human_T_beta", recomb_type="VDJ")
    
    n = 1000
    df = pd.read_csv('cdr3_beta_500K.csv')
    inputlist = df.iloc[:,0].to_list()[0:n]
    inputlist  = [x for x in inputlist if _valid_cdr3(x)]

    @timer
    def olga_in_series(f = olga_beta.compute_aa_cdr3_pgen, input = inputlist):
        return [f(x) for x in input]

    @timer
    def olga_in_parmap(f = olga_beta.compute_aa_cdr3_pgen, input = inputlist, **kwargs):
        return parmap.map(f, input, pm_pbar=True, **kwargs)

    r1 = olga_in_series(f = olga_beta.compute_aa_cdr3_pgen, input = inputlist)
    r2 = olga_in_parmap(f = olga_beta.compute_aa_cdr3_pgen, input = inputlist)
    
    assert np.all(r1 == r2)


def test_pgen_with_parmap():
    """
    Really simple example of using multiple cpus to 
    speed up computation of pgens with olga.
    """
    import parmap
    from tcrdist.pgen import OlgaModel
    olga_beta  = OlgaModel(chain_folder = "human_T_beta", recomb_type="VDJ")
    parmap.map(olga_beta.compute_aa_cdr3_pgen, ['CASSYRVGTDTQYF', 'CATSTNRGGTPADTQYF', 'CASQGDSFNSPLHF', 'CASSPWTGSMALHF'])


import pytest

def test_alpha_beta():
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory2
    from tcrdist.rep_funcs import  compute_n_tally_out_of_memory2
    from hierdiff.association_testing import cluster_association_test

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df.sample(100, random_state = 1), 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv', 
                compute_distances = True,
                store_all_cdr = False)

    check_beta = tr.pw_beta.copy(); check_beta[check_beta == 0] = 1
    check_alpha = tr.pw_alpha.copy(); check_alpha[check_alpha == 0] = 1
    check_alpha_beta = check_beta + check_alpha
    

    S, fragments = compute_pw_sparse_out_of_memory2(    tr = tr,
                                                        row_size      = 50,
                                                        pm_processes  = 1,
                                                        pm_pbar       = True,
                                                        max_distance  = 1000,
                                                        reassemble    = True,
                                                        cleanup       = False,
                                                        assign        = True)
    
    assert np.all(tr.pw_beta == check_beta)
    assert np.all(tr.pw_alpha == check_alpha)

    ndif1 = compute_n_tally_out_of_memory2(fragments, 
                                         to_file = False, 
                                         to_memory = True,
                                         pm_processes = 2, 
                                         x_cols = ['epitope'],
                                         count_col='count',
                                         knn_neighbors= None,
                                         knn_radius =100)

    from hierdiff.association_testing import cluster_association_test
    ndif1 = cluster_association_test(res = ndif1, y_col='cmember', method='chi2')


    from tcrdist.rep_diff import neighborhood_diff
    ndif2 = neighborhood_diff(clone_df= tr.clone_df, 
        pwmat = np.array(tr.pw_beta.todense() + tr.pw_alpha.todense()),
        count_col = 'count', 
        x_cols = ['epitope'], 
        knn_radius = 100, 
        test_method = "chi2")

    assert ndif1.shape == ndif2.shape
    np.all(ndif2['FDRq'].to_list() == ndif2['FDRq'].to_list())





def test_beta():
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory2
    from tcrdist.rep_funcs import  compute_n_tally_out_of_memory2
    from hierdiff.association_testing import cluster_association_test

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df.sample(100, random_state = 1), 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv', 
                compute_distances = True,
                store_all_cdr = False)

    check_beta = tr.pw_beta.copy(); check_beta[check_beta == 0] = 1

    S, fragments = compute_pw_sparse_out_of_memory2(    tr = tr,
                                                        row_size      = 50,
                                                        pm_processes  = 2,
                                                        pm_pbar       = True,
                                                        max_distance  = 1000,
                                                        reassemble    = True,
                                                        cleanup       = False,
                                                        assign        = True)
    import numpy as np
    assert np.all(tr.pw_beta == check_beta)
    assert 'pw_alpha' not in tr.__dict__.keys()


    ndif1 = compute_n_tally_out_of_memory2(fragments, 
                                         to_file = False, 
                                         to_memory = True,
                                         pm_processes = 2, 
                                         x_cols = ['epitope'],
                                         count_col='count',
                                         knn_neighbors= None,
                                         knn_radius =100)

    from hierdiff.association_testing import cluster_association_test
    ndif1 = cluster_association_test(res = ndif1, y_col='cmember', method='chi2')


    from tcrdist.rep_diff import neighborhood_diff
    ndif2 = neighborhood_diff(clone_df= tr.clone_df, 
        pwmat = np.array(tr.pw_beta.todense()),
        count_col = 'count', 
        x_cols = ['epitope'], 
        knn_radius = 100, 
        test_method = "chi2")

    assert ndif1.shape == ndif2.shape
    np.all(ndif2['FDRq'].to_list() == ndif2['FDRq'].to_list())



def test_alpha():
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory2
    from tcrdist.rep_funcs import  compute_n_tally_out_of_memory2
    from hierdiff.association_testing import cluster_association_test

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df.sample(100, random_state = 1), 
                organism = 'mouse', 
                chains = ['alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv', 
                compute_distances = True,
                store_all_cdr = False)

    check_alpha = tr.pw_alpha.copy(); check_alpha[check_alpha == 0] = 1

    S, fragments = compute_pw_sparse_out_of_memory2(    tr = tr,
                                                        row_size      = 50,
                                                        pm_processes  = 2,
                                                        pm_pbar       = True,
                                                        max_distance  = 1000,
                                                        reassemble    = True,
                                                        cleanup       = False,
                                                        assign        = True)
    import numpy as np
    assert np.all(tr.pw_alpha == check_alpha)
    assert 'pw_beta' not in tr.__dict__.keys()

    ndif1 = compute_n_tally_out_of_memory2(fragments, 
                                         to_file = False, 
                                         to_memory = True,
                                         pm_processes = 2, 
                                         x_cols = ['epitope'],
                                         count_col='count',
                                         knn_neighbors= None,
                                         knn_radius =100)

    from hierdiff.association_testing import cluster_association_test
    ndif1 = cluster_association_test(res = ndif1, y_col='cmember', method='chi2')


    from tcrdist.rep_diff import neighborhood_diff
    ndif2 = neighborhood_diff(clone_df= tr.clone_df, 
        pwmat = np.array(tr.pw_alpha.todense()),
        count_col = 'count', 
        x_cols = ['epitope'], 
        knn_radius = 100, 
        test_method = "chi2")

    assert ndif1.shape == ndif2.shape
    np.all(ndif2['FDRq'].to_list() == ndif2['FDRq'].to_list())




from scipy import sparse
import numpy as np

def add_sparse_pwd(a, b, neg1_fmt=True):
    """
    Add sparse pairwise distance matrices a and b.
    Can work with two "formats": true zero distances as -1
    or true zero distances as 0. Returns same format as received.
    Does not attempt to add sparse zeros (ie NA),
    so returns an intersection of data in a and b (summed).
    Returns a scipy.sparse.csr matrix (can take csr or any
    format as inputs).
    Parameters
    ----------
    a, b : scipy.sparse.csr or coo
        Sparse pairwise distance matrices. D=0 can be 
        represented as -1 or 0 but should be specified
        using neg1_fmt
    neg1_fmt : bool
        Indicator whether to assume D=0 is represented as -1
    Returns
    -------
    tot : scipy.sparse.csr
        Sparse matrix representing sum of a and b
        where a and b both have values.
    """
    
    a_coo = a.tocoo()
    b_coo = b.tocoo()
    nr, nc = a_coo.shape
    
    ind_a = a_coo.row.astype(np.int64) * nc + a_coo.col
    ind_b = b_coo.row.astype(np.int64) * nc + b_coo.col
    # ind_ab = np.array(tuple(set(ind_a).intersection(set(ind_b))))
    ind_ab = np.intersect1d(ind_a, ind_b, assume_unique=True)
    keep_a = np.in1d(ind_a, ind_ab, assume_unique=True)
    keep_b = np.in1d(ind_b, ind_ab, assume_unique=True)
    if neg1_fmt:
        a_data = a_coo.data[keep_a]
        a_data[a_data == -1] = 0
        b_data = b_coo.data[keep_b]
        b_data[b_data == -1] = 0
        tot = a_data + b_data
    else:
        tot = a_coo.data[keep_a] + b_coo.data[keep_b]
    """Need to add -1 as placeholder upon creation of new sparse matrix"""
    tot[tot == 0] = -1
    aplusb = sparse.coo_matrix((tot,
                                (a_coo.row[keep_a], a_coo.col[keep_a])), shape=(nr, nc))
    aplusb = aplusb.tocsr()
    if not neg1_fmt:
        """Convert it back to 0 = true zero distance"""
        aplusb.data[aplusb.data == -1] = 0
    return aplusb



import sys
import os

path_to_scripts = os.path.dirname(os.path.realpath(__file__))
path_to_db = os.path.join(path_to_scripts, 'db')
path_to_default_models = os.path.join(path_to_scripts, 'default_models')
path_to_base = os.path.dirname(path_to_scripts)
path_to_data = os.path.join(path_to_scripts, 'data')



import olga.load_model as load_model
import olga.generation_probability as pgen
import olga.sequence_generation as seq_gen
import pandas as pd


def generate_simulated_beta_seqs(params_file_name = 'tcrdist/default_models/human_T_beta/model_params.txt',
                            marginals_file_name = 'tcrdist/default_models/human_T_beta/model_marginals.txt',
                            V_anchor_pos_file ='tcrdist/default_models/human_T_beta/V_gene_CDR3_anchors.csv',
                            J_anchor_pos_file = 'tcrdist/default_models/human_T_beta/J_gene_CDR3_anchors.csv',
                            output_cols = ['cdr3_b_aa', "v_b_gene",'j_b_gene'],
                            n = 100000):
    #Load data
    genomic_data = load_model.GenomicDataVDJ()
    genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
    #Load model
    generative_model = load_model.GenerativeModelVDJ()
    generative_model.load_and_process_igor_model(marginals_file_name)
    seq_gen_model = seq_gen.SequenceGenerationVDJ(generative_model, genomic_data)

    #Generate some random sequences


    vs=[x[0] for x in genomic_data.__dict__['genV']]
    js=[x[0] for x in genomic_data.__dict__['genJ']]
    vs = {i:k for i,k in enumerate(vs)}
    js = {i:k for i,k in enumerate(js)}

    sim_cdr3 = [seq_gen_model.gen_rnd_prod_CDR3()[1:4] for x in range(n)]
    sim_cdr3_long = [(i,vs[v],js[j]) for i,v,j in sim_cdr3 ]

    df = pd.DataFrame(sim_cdr3_long, columns = output_cols)
    return df

def generate_simulated_alpha_seqs(params_file_name = 'tcrdist/default_models/human_T_alpha/model_params.txt',
                            marginals_file_name = 'tcrdist/default_models/human_T_alpha/model_marginals.txt',
                            V_anchor_pos_file ='tcrdist/default_models/human_T_alpha/V_gene_CDR3_anchors.csv',
                            J_anchor_pos_file = 'tcrdist/default_models/human_T_alpha/J_gene_CDR3_anchors.csv',
                            output_cols = ['cdr3_a_aa', "v_a_gene",'j_a_gene'],
                            n = 100000):
    #Load data
    genomic_data = load_model.GenomicDataVJ()
    genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
    #Load model
    generative_model = load_model.GenerativeModelVJ()
    generative_model.load_and_process_igor_model(marginals_file_name)
    seq_gen_model = seq_gen.SequenceGenerationVJ(generative_model, genomic_data)

    #Generate some random sequences
    vs=[x[0] for x in genomic_data.__dict__['genV']]
    js=[x[0] for x in genomic_data.__dict__['genJ']]
    vs = {i:k for i,k in enumerate(vs)}
    js = {i:k for i,k in enumerate(js)}

    sim_cdr3 = [seq_gen_model.gen_rnd_prod_CDR3()[1:4] for x in range(n)]
    sim_cdr3_long = [(i,vs[v],js[j]) for i,v,j in sim_cdr3 ]

    df = pd.DataFrame(sim_cdr3_long, columns = output_cols)
    return df

if __name__ == "__main__":
    """
    Using Olga See: 
    ---------------
    Zachary Sethna, Yuval Elhanati, Curtis G Callan, Aleksandra M Walczak, Thierry Mora
    `Bioinformatics (2019) <https://doi.org/10.1093/bioinformatics/btz035>`_ 
    OLGA: fast computation of generation probabilities of B- and T-cell receptor amino acid sequences and motifs


    Generate 1000K (1M) CDR3s using default Olga Models
    Human (Alpha/Beta) and Mouse (Beta)

    human_T_alpha_sim1000K.csv
    human_T_beta_sim1000K.csv
    mouse_T_beta_sim1000K.csv
    
    contained in: 
    olga_T_alpha_beta_1000K_simulated_cdr3.zip 
    """
    dfb= generate_simulated_beta_seqs(params_file_name = 'tcrdist/default_models/human_T_beta/model_params.txt',
                                marginals_file_name = 'tcrdist/default_models/human_T_beta/model_marginals.txt',
                                V_anchor_pos_file ='tcrdist/default_models/human_T_beta/V_gene_CDR3_anchors.csv',
                                J_anchor_pos_file = 'tcrdist/default_models/human_T_beta/J_gene_CDR3_anchors.csv',
                                output_cols = ['cdr3_b_aa', "v_b_gene",'j_b_gene'], n = 1000000)
    dfb.to_csv('human_T_beta_sim1000K.csv', index = False)

    dfa = generate_simulated_alpha_seqs(params_file_name = 'tcrdist/default_models/human_T_alpha/model_params.txt',
                                marginals_file_name = 'tcrdist/default_models/human_T_alpha/model_marginals.txt',
                                V_anchor_pos_file ='tcrdist/default_models/human_T_alpha/V_gene_CDR3_anchors.csv',
                                J_anchor_pos_file = 'tcrdist/default_models/human_T_alpha/J_gene_CDR3_anchors.csv',
                                output_cols = ['cdr3_a_aa', "v_a_gene",'j_a_gene'],
                                n = 1000000)

    dfa.to_csv('human_T_alpha_sim1000K.csv', index = False)                         

    dfb= generate_simulated_beta_seqs(params_file_name = 'tcrdist/default_models/mouse_T_beta/model_params.txt',
                                marginals_file_name = 'tcrdist/default_models/mouse_T_beta/model_marginals.txt',
                                V_anchor_pos_file ='tcrdist/default_models/mouse_T_beta/V_gene_CDR3_anchors.csv',
                                J_anchor_pos_file = 'tcrdist/default_models/mouse_T_beta/J_gene_CDR3_anchors.csv',
                                output_cols = ['cdr3_b_aa', "v_b_gene",'j_b_gene'], n = 1000000)
    dfb.to_csv('mouse_T_beta_sim1000K.csv', index = False)



import numpy as np
import pandas as pd
import copy
import os
import parmap 
import secrets
from progress.bar import IncrementalBar
from hierdiff.tally import neighborhood_tally
import pwseqdist


from scipy import sparse
"""
Large datasets can consume more memory than is typically available. 

The memory module solves this problem with functions
that allow for TCRdistance computation to 
be broken down into fragments that can be stored on disk
and reassembled in a sparser format. 
"""

def _partition(l, n):
    """
    _partition breaks up a list in parts of size n

    Parameters
    ----------
    l : list
    
    n : int

    >>> _partition([1,2,3,4,5,6], 2)
    [[1, 2], [3, 4], [5, 6]]
    >>> _partition([1,2,3,4,5,6], 3)
    [[1, 2, 3], [4, 5, 6]]
        """
    n = int(n)
    r = [l[i:i + n] for i in range(0, len(l), n)]
    return r


def gen_sparse_rw_on_fragment(tcrrep, ind, outfile, matrix_name = 'rw_beta', max_distance=50):
    """
    gen_sparse_rw_on_fragment generates a sparse matrix of distances
    on a fragment of overall comparisons. 

    Suppose a clone_df has m unique clones. A full matrix is m x m. 
    this computes distances of the rectangular fragment 
    m[ind,] x m. The fragment is converted to a sparse compressed 
    format. All true zeros are set to 1, and all values above <max_distance>
    are set to one, for a massive space savings.

    tcrrep : TCRrep
        TCRrep instance 
    ind : list
        line of row index position specifying the rows of the clone_df
    outfile : str
        string name of the fragment
    matrix_name  : str
        the matrix attribute to be stored ('rw_beta')

    Example
    -------
    See usage in tcrdist.rep_funcs.compute_pw_sparse_out_of_memory

    """
    tr = copy.deepcopy(tcrrep)
    tr.compute_rect_distances(df = tr.clone_df.iloc[ind,], df2 = tr.clone_df, store = False)
    M = getattr(tr, matrix_name)
    # Set all true 0s to -1
    M[M == 0] = -1
    # Set all values greater than max_distance to zero to create massive scarcity
    M[M > max_distance] = 0 
    # convert to scr matrix spa
    M = sparse.csr_matrix(M)
    # Write to outfile
    sparse.save_npz(outfile, M)
    del tr
    del M
    return True

def gen_sparse_rw_on_fragment2(tcrrep, ind, outfile, max_distance=50):
    """
    gen_sparse_rw_on_fragment generates a sparse matrix of distances
    on a fragment of overall comparisons. 

    Suppose a clone_df has m unique clones. A full matrix is m x m. 
    this computes distances of the rectangular fragment 
    m[ind,] x m. The fragment is converted to a sparse compressed 
    format. All true zeros are set to 1, and all values above <max_distance>
    are set to one, for a massive space savings.

    tcrrep : TCRrep
        TCRrep instance 
    ind : list
        line of row index position specifying the rows of the clone_df
    outfile : str
        string name of the fragment
    matrix_name  : str
        the matrix attribute to be stored ('rw_beta')

    Example
    -------
    See usage in tcrdist.rep_funcs.compute_pw_sparse_out_of_memory

    """
    tr = copy.deepcopy(tcrrep)
    tr.compute_rect_distances(df = tr.clone_df.iloc[ind,], df2 = tr.clone_df, store = False)
    for chain in tr.chains:
        M = getattr(tr, f"rw_{chain}")
        # Set all true 0s to 1s
        M[M == 0] = 1
        # Set all values greater than max_distance to zero to create massive scarcity
        M[M > max_distance] = 0 
        # convert to scr matrix spa
        M = sparse.csr_matrix(M)
        # Write to outfile
        x = f"rw_{chain}"
        sparse.save_npz(f"{outfile}.{x}.npz", M)
        del M
    
    del tr
    return True


def collapse_csrs(list_of_csr_filenames, axis = 0):
    """
    Given a list of filenames referring to the sparse
    representation of ordered row-wise (axis = 0) chunks
    of a matrix, collapse chunks into a single 
    matrix in sparse format.

    Parameters
    ----------
    list_of_csr_filenames : list of strings
        list of string pointing to filenames of .npz sparse matrices
    axis : int
        if 0, combine with vstack, else if 1

    Example
    -------
    >>> import numpy as np
    >>> from scipy import sparse
    >>> A = np.array([[1, 0, 0, 1, 0, 0], [0, 0, 2, 0, 0, 1], [0, 0, 0, 2, 0, 0]])
    >>> B = np.array([[3, 3, 3, 1, 0, 0], [5, 5, 2, 0, 0, 1], [6, 7, 0, 2, 0, 0]])
    >>> A_s = sparse.csr_matrix(A)
    >>> B_s = sparse.csr_matrix(B)
    >>> sparse.save_npz("A.npz", A_s)
    >>> sparse.save_npz("B.npz", B_s)
    >>> AB_s = collapse_csrs(["A.npz","B.npz"], axis = 0)
    >>> AB = AB_s.todense()
    >>> assert np.all(AB == np.concatenate([A,B], axis = 0))
    """
    assert axis in [0,1], "axis must be 0 (for hstack on rows) or 1 (for vstack on columns)."
    if axis == 0:
        full_spase_matrix = sparse.vstack([sparse.load_npz(f_npz) for f_npz in list_of_csr_filenames])
    if axis == 1:
        full_spase_matrix = sparse.hstack([sparse.load_npz(f_npz) for f_npz in list_of_csr_filenames])
    return full_spase_matrix



def gen_n_tally_on_fragment(tcrrep, 
                            ind, 
                            infile, 
                            outfile,
                            x_cols = ['epitope'], 
                            count_col='count', 
                            knn_neighbors= None, 
                            knn_radius =50):
    import copy
    tr = copy.deepcopy(tcrrep)
    #with open(infile ,"rb") as frag:
    #   rwmat = np.load(frag)
    rwmat = sparse.load_npz(infile)
    rwmat = np.asarray(rwmat.todense())
    rwmat[rwmat == 0] = 500
    ndif = neighborhood_tally(  df_pop = tr.clone_df, 
                                pwmat = rwmat,#tr.pw_beta[ind,], 
                                x_cols = x_cols, 
                                df_centroids=tr.clone_df.iloc[ind,],
                                count_col=count_col, 
                                knn_neighbors= knn_neighbors, 
                                knn_radius =knn_radius)

    ndif.to_csv(outfile, index = False)
    
    del ndif
    
    return outfile


def gen_n_tally_on_fragment2(tcrrep, 
                            ind, 
                            infile, 
                            outfile,
                            x_cols = ['epitope'], 
                            count_col='count', 
                            knn_neighbors= None, 
                            knn_radius =50):
    import copy
    tr = copy.deepcopy(tcrrep)
    
    rwmat_dict = dict()
    for chain in tr.chains:
        rwmat = sparse.load_npz(f"{infile}.rw_{chain}.npz")
        rwmat = np.asarray(rwmat.todense())
        rwmat[rwmat == 0] = 500
        rwmat_dict[chain] = rwmat

    if len(tr.chains) == 1:
        rwmat = rwmat_dict[tr.chains[0]]
    if len(tr.chains) == 2:
        rwmat = np.add(rwmat_dict[tr.chains[0]], rwmat_dict[tr.chains[1]])
        
    ndif = neighborhood_tally(  df_pop = tr.clone_df, 
                                pwmat = rwmat,#tr.pw_beta[ind,], 
                                x_cols = x_cols, 
                                df_centroids=tr.clone_df.iloc[ind,],
                                count_col=count_col, 
                                knn_neighbors= knn_neighbors, 
                                knn_radius =knn_radius)

    ndif.to_csv(outfile, index = False)
    
    del ndif
    
    return outfile


def _concat_to_file(dest, fragments, filename = "nndif.csv" ):
    """
    _concat_to_file take a list of .csv filenames, <fragments>. 
    and combines them into one concatenated file 
    written written to the <dest> directory on disk.

    Parameters
    ----------
    dest : str
        location where file is to be written 
    fragments : list
        list of .csv files
    Returns
    -------
    concatenated_filename : str
        filename of the concatenated file
    """
    from progress.bar import IncrementalBar
    bar = IncrementalBar(f'Processing Files', max = len(fragments), suffix='%(percent)d%%')
    mysep = ","
    concatenated_filename = os.path.join(dest, filename)
    counter = 0
    for f in fragments:
        df = pd.read_csv(os.path.join(f), sep = ',')
        if counter == 0:
            df.to_csv(concatenated_filename, header=True, index = False, sep = mysep)
        if counter > 0:
            df.to_csv(concatenated_filename, header=False, index = False, sep = mysep, mode = "a")
        del df
        counter =+1 
        bar.next()
    bar.next(); bar.finish()
    print(f"WROTE TO {concatenated_filename}")
    return concatenated_filename
    

def _concat_to_memory(fragments):
    """
    This
    Parameters
    ----------
    fragments : list
        list of .csv files
    Returns
    -------
    concatenated_df : DataFrame
        
    """
    from progress.bar import IncrementalBar
    bar = IncrementalBar(f'Processing Files', max = len(fragments), suffix='%(percent)d%%')
    mysep = ","
    dfs = list()
    for f in fragments:
        dfs.append(pd.read_csv(os.path.join(f), sep = ','))
        bar.next()
    concatenated_df = pd.concat(dfs)
    bar.finish()
    return concatenated_df


def _sparse_cdr3_tcrdist_shard(chunk1_indices, components, df, metrics, weights, kargs, radius, df2):
    """Returns a sparse matrix with 0/NA meaning D > radius and -1 meaning D = 0"""
    
    """For now not trying to re-expand less as I expand components and find D > radius"""
    # keep_ind = np.ones(chunk1_indices.shape[0], dtype=bool)

    if df2 is None:
        n2 = df.shape[0]
    else:
        n2 = df2.shape[0]

    metric_keys = [k for k in metrics.keys() if not 'cdr3' in k]
    
    metric_cdr3_keys = [k for k in metrics.keys() if 'cdr3' in k]
    weight_cdr3_keys = [k for k in weights.keys() if 'cdr3' in k]
    assert metric_cdr3_keys == weight_cdr3_keys, "metrics and weights cdr3 keys must be identical"
    
    if kargs is not None:
        kargs_cdr3_keys  = [k for k in kargs.keys() if 'cdr3' in k]
        assert metric_cdr3_keys == kargs_cdr3_keys,  "metrics and kargs cdr3 keys must be identical"

    tcrdist = np.zeros((chunk1_indices.shape[0], n2), dtype=np.int16)
    for k in metric_keys:
        """Re-expand chunk of rows and all of columns: weights already applied"""
        tcrdist = tcrdist + components[k][0][components[k][1][chunk1_indices], :][:, components[k][2]]

    """Make a list of pairs to compute CDR3 distance for this chunk.
    Then create sparse CSR matrix for the chunk and return"""
    for k in metric_cdr3_keys:
        """Compute CDR3s sparsely"""
        """Only compute CDR3 distance for if the running total <= radius:
        Not currently possible with running rect and sparse doesn't avoid redundant distances. Opting
        for computing the whole chunk with pwrect"""

        """Since this is just a chunk i could use regular pwrect and not worry about running.
        Running won't be very helpful since it won't incorporate the radius from the other components"""
        # apply_running_rect(metric, seqs1, seqs2, radius=radius, density_est=density_est, *args, ncpus=1, uniqify=True, alphabet=parasail_aa_alphabet, **kwargs)
        if df2 is None:
            seqs2 = df[k].values
        else:
            seqs2 = df2[k].values
        pwrect = pwseqdist.apply_pairwise_rect(metric=metrics[k], 
                                        seqs1=df[k].values[chunk1_indices], 
                                        seqs2=seqs2, 
                                        ncpus=1, 
                                        uniqify=True, 
                                        reexpand=True,
                                        **kargs[k])
        tcrdist = tcrdist + weights[k] * pwrect
    zind = np.nonzero(tcrdist == 0)
    tcrdist[zind] = -1
    keep_ind = np.nonzero(tcrdist <= radius)
    S = sparse.csr_matrix((tcrdist[keep_ind], keep_ind), shape=tcrdist.shape)
    return S


import pytest
import pandas as pd
v20df = pd.read_csv('https://www.dropbox.com/s/wmc5wc752t782kq/MIRA_v21_covid_diagnosed_sars_cov2_ci_epitope_specific_tcrs.tsv?dl=1', sep = "\t").head(1000)
v21df = pd.read_csv('https://www.dropbox.com/s/c3gfq1lu0xdefpy/MIRA_v20_covid_diagnosed_sars_cov2_ci_epitope_specific_tcrs.tsv?dl=1', sep = "\t").head(1000)

"""
Example 1, Edit Dist 1 join
"""
def test_tcr_join_edit1():
  import pandas as pd
  from tcrdist.join import join_by_dist
  from tcrdist.breadth import get_safe_chunk
  from tcrdist.rep_funcs import compute_pws_sparse
  import pwseqdist as pw

  my_metrics = { "cdr3_b_aa" : pw.metrics.nb_vector_editdistance}
  my_weights = { "cdr3_b_aa" : 1}
  my_kargs = {"cdr3_b_aa" :{'use_numba': True}}
  distances = compute_pws_sparse(
      df= v21df,
      df2 = v20df ,
      metrics = my_metrics,
      weights = my_weights,
      kargs = my_kargs,
      radius=1,
      cpu=2, 
      chunk_size=get_safe_chunk(v21df.shape[0], v20df.shape[0]), 
      store=False, 
      pm_pbar=True)
  csrmat = distances['tcrdist'] # it's called a tcrdist here, but was computed as edit distance 1
  left_join_df = join_by_dist(
      how = 'left',
      csrmat = csrmat,
      left_df = v21df,
      right_df = v20df,
      left_cols  = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      right_cols = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      left_suffix = '_x',
      right_suffix = '_y',
      max_n= 10,
      radius = 1)
  inner_join_df = join_by_dist(
      how = 'inner',
      csrmat = csrmat,
      left_df = v21df,
      right_df = v20df,
      left_cols  = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      right_cols = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      left_suffix = '_x',
      right_suffix = '_y',
      max_n= 10,
      radius = 1)
  outer_join_df = join_by_dist(
      how = 'outer',
      csrmat = csrmat,
      left_df = v21df,
      right_df = v20df,
      left_cols  = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      right_cols = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      left_suffix = '_x',
      right_suffix = '_y',
      max_n= 10,
      radius = 1)
  
  assert left_join_df.shape[0] > inner_join_df.shape[0]
  assert outer_join_df.shape[0] > inner_join_df.shape[0]
  assert outer_join_df.shape[0] > left_join_df.shape[0]
    
"""
2. Full Example using TCRdist on all CDRs on real data from two MIRA cohorts, 
Here we use TCRrep to infer CDR1,2,2.5 to compute a full tcrdist 
"""
def test_tcr_join_tcrdist():
  import pandas as pd
  from tcrdist.breadth import get_safe_chunk
  from tcrdist.repertoire import TCRrep
  from tcrdist.join import join_by_dist

  tr20 = TCRrep(cell_df = v20df[['subject', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'bio_identity','protein_coordinate']].copy(),
      organism='human', 
      chains=['beta'], 
      compute_distances = False)
  tr21 = TCRrep(cell_df = v21df[['subject', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'bio_identity', 'protein_coordinate']].copy(),
      organism='human', 
      chains=['beta'], 
      compute_distances = False)
  tr21.cpus = 2
  tr21.compute_sparse_rect_distances(df = tr21.clone_df, df2 = tr20.clone_df, radius = 36, chunk_size = get_safe_chunk(tr21.clone_df.shape[0], tr20.clone_df.shape[0]))

  left_right_comparision = join_by_dist(
      how = 'inner',
      csrmat = tr21.rw_beta,
      left_df = v21df,
      right_df = v20df,
      left_cols  = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      right_cols = ['cdr3_b_aa','v_b_gene','j_b_gene','protein_coordinate','bio_identity','subject'],
      left_suffix = '_x',
      right_suffix = '_y',
      max_n= 10,
      radius = 24)




import pytest
"""OLGA (Sethna et al., 2019) is a fast tool for estimating the probability of generations
(i.e. Pgen). In this example we are going to estimate Pgens for the dash.csv data.
"""
def test_pgen_mouse():
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.pgen import OlgaModel
    import numpy as np

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True)

    # Load OLGA model as a python object
    olga_beta  = OlgaModel(chain_folder = "mouse_T_beta", recomb_type="VDJ")
    olga_alpha = OlgaModel(chain_folder = "mouse_T_alpha", recomb_type="VJ")

    # An example computing a single Pgen
    olga_beta.compute_aa_cdr3_pgen(tr.clone_df['cdr3_b_aa'][0])
    olga_alpha.compute_aa_cdr3_pgen(tr.clone_df['cdr3_a_aa'][0])

    # An example computing multiple Pgens 
    olga_beta.compute_aa_cdr3_pgens(tr.clone_df['cdr3_b_aa'][0:5])
    olga_alpha.compute_aa_cdr3_pgens(tr.clone_df['cdr3_a_aa'][0:5])

    # An example computing 1920 Pgens more quickly with multiple cpus
    import parmap
    tr.clone_df['pgen_cdr3_b_aa'] = \
        parmap.map(
            olga_beta.compute_aa_cdr3_pgen, 
            tr.clone_df['cdr3_b_aa'], 
            pm_pbar=True, 
            pm_processes = 2)

    tr.clone_df['pgen_cdr3_a_aa'] = \
        parmap.map(
            olga_alpha.compute_aa_cdr3_pgen, 
            tr.clone_df['cdr3_a_aa'], 
            pm_pbar=True, 
            pm_processes = 2)

    """
    We can do something else useful. We've tweaked the original 
    generative code in OLGA, so that you can generate CDRs,
    given a specific TRV and TRJ. 

    Note that unfortunately not all genes are recognized in default OLGA models, 
    but many are. This gives you an idea of what you can do. Here are 10
    CDR3s generated at random given a particular V,J usage combination
    """
    np.random.seed(1)
    olga_beta.gen_cdr3s(V = 'TRBV14*01', J = 'TRBJ2-5*01', n =10) 
    olga_alpha.gen_cdr3s(V ='TRAV4-3*02', J ='TRAJ31*01',  n =10)

    """
    Using this approach, we can synthesize an 100K background, 
    with similar gene usage frequency to our actual repertoire. 
    Note, however, that given data availability, 
    this is currently likely the most reliable for human beta chain.

    After OLGA's publication, a default mouse alpha model (mouse_T_alpha) 
    was added to the OLGA GitHub repository. We've included that here
    but it should be used with caution as it is missing
    a number of commonly seen V genes.
    """
    np.random.seed(1)
    tr.synthesize_vj_matched_background(chain = 'beta')
    """
              v_b_gene    j_b_gene            cdr3_b_aa        pV        pJ       pVJ   weights      source
    0        TRBV14*01  TRBJ2-3*01        CASSLASAETLYF  0.033721  0.092039  0.002989  0.065742  vj_matched
    1      TRBV13-2*01  TRBJ2-3*01    CASGDAPDRTGAETLYF  0.118785  0.092039  0.010331  0.271309  vj_matched
    2      TRBV13-3*01  TRBJ1-1*01  CASSDGFSRTGGVNTEVFF  0.074051  0.106146  0.006923  1.009124  vj_matched
    3      TRBV13-3*01  TRBJ2-1*01       CASSDVQGGAEQFF  0.074051  0.117684  0.008915  1.021244  vj_matched
    4      TRBV13-3*01  TRBJ2-7*01     CASSSGTGGYIYEQYF  0.074051  0.204898  0.015366  1.670224  vj_matched
    ...            ...         ...                  ...       ...       ...       ...       ...         ...
    99995    TRBV14*01  TRBJ2-3*01  CASSPTGGAPYASAETLYF  0.033721  0.092039  0.002989  0.065742  vj_matched
    99996    TRBV17*01  TRBJ2-5*01       CASSRDPTQDTQYF  0.028110  0.124712  0.004930  0.650360  vj_matched
    99997    TRBV14*01  TRBJ2-3*01       CASSSTGGAETLYF  0.033721  0.092039  0.002989  0.065742  vj_matched
    99998  TRBV13-1*01  TRBJ2-1*01      CASSDWGKDYAEQFF  0.106042  0.117684  0.013373  2.622194  vj_matched
    99999     TRBV4*01  TRBJ2-3*01      CASSYDRGSAETLYF  0.040749  0.092039  0.002989  0.068343  vj_matched
    """
    np.random.seed(1)
    tr.synthesize_vj_matched_background(chain = 'alpha')
    """
               v_a_gene   j_a_gene        cdr3_a_aa        pV        pJ       pVJ   weights      source
    0      TRAV12N-3*01  TRAJ34*02     CAIASNTNKVVF  0.000438  0.000088  0.000088  0.006059  vj_matched
    1       TRAV3D-3*02  TRAJ33*01  CAVSAGADSNYQLIW  0.000088  0.000088  0.000088  0.005122  vj_matched
    2        TRAV3-3*01  TRAJ27*01     CAVSTNTGKLTF  0.014029  0.042964  0.000877  0.277471  vj_matched
    3        TRAV3-3*01  TRAJ26*01    CAVSHNYAQGLTF  0.014029  0.040947  0.001052  0.009155  vj_matched
    4        TRAV3-3*01  TRAJ26*01   CAVSARNYAQGLTF  0.014029  0.040947  0.001052  0.009155  vj_matched
    ...             ...        ...              ...       ...       ...       ...       ...         ...
    99995   TRAV3D-3*02  TRAJ21*01    CAVSVSNYNVLYF  0.000088  0.039982  0.000088  0.003758  vj_matched
    99996    TRAV3-3*01  TRAJ43*01    CAVSENNNNAPRF  0.014029  0.022271  0.000526  0.071093  vj_matched
    99997   TRAV3D-3*02  TRAJ26*01    CAVSGNYAQGLTF  0.000088  0.040947  0.000088  0.000296  vj_matched
    99998    TRAV3-3*01  TRAJ26*01   CAVKGNNYAQGLTF  0.014029  0.040947  0.001052  0.009155  vj_matched
    99999   TRAV9N-2*01  TRAJ15*01      CTYQGGRALIF  0.000088  0.043840  0.000088  0.020438  vj_matched
    """

    """"
    tcrdist3's integration of Pgen estimates makes it very easy to look for
    PUBLIC clusters of TCRs (i.e. high number of neighbors) with unlikely V(D)J 
    recombinations.
    """
    from tcrdist.public import _neighbors_fixed_radius
    from tcrdist.public import _K_neighbors_fixed_radius
    tr.clone_df['neighbors'] = _neighbors_fixed_radius(pwmat = tr.pw_beta, radius = 18)
    tr.clone_df['K_neighbors'] = _K_neighbors_fixed_radius(pwmat = tr.pw_beta , radius = 18)
    tr.clone_df['pgen_cdr3_b_aa_nlog10'] = tr.clone_df['pgen_cdr3_b_aa'].apply(lambda x : -1*np.log10(x))
    tr.clone_df['nsubject'] = tr.clone_df['neighbors'].apply(lambda x: len(tr.clone_df['subject'][x].unique()))
    # nsubject > 1 implies quasi-publicity
    tr.clone_df['qpublic'] = tr.clone_df['nsubject'].apply(lambda x: x >1)

    # Note one can find neighbors based on paired-chain distances.
    from tcrdist.public import _neighbors_fixed_radius
    from tcrdist.public import _K_neighbors_fixed_radius
    tr.clone_df['neighbors'] = _neighbors_fixed_radius(pwmat = tr.pw_beta + tr.pw_alpha, radius = 50)
    tr.clone_df['K_neighbors'] = _K_neighbors_fixed_radius(pwmat = tr.pw_beta + tr.pw_alpha , radius = 50)
    tr.clone_df['pgen_cdr3_b_aa_nlog10'] = tr.clone_df['pgen_cdr3_b_aa'].apply(lambda x : -1*np.log10(x))
    tr.clone_df['nsubject'] = tr.clone_df['neighbors'].apply(lambda x: len(tr.clone_df['subject'][x].unique()))
    # nsubject > 1 implies quasi-publicity)
    tr.clone_df['qpublic'] = tr.clone_df['nsubject'].apply(lambda x: x >1 )
    """
    #Code for plotting pgen vs. neighbors 
    import plotnine
    from plotnine import ggplot, geom_point, aes, ylab, xlab
    (ggplot(tr.clone_df, 
        aes(x= 'pgen_cdr3_b_aa_nlog10', 
            y ='K_neighbors',
            color = 'nsubject')) + 
        geom_point(size = .1) + 
        plotnine.scales.xlim(0,20) + 
        plotnine.facets.facet_wrap('epitope', scales = "free_y")+ ylab('Neighbors') + xlab('-Log10 Pgen'))
    """